# -*- coding: utf-8 -*-
"""
This module provides support for closures and inner functions.

@autojit
def outer():
    a = 10 # this is a cellvar

    @jit('void()')
    def inner():
        print a # this is a freevar

    inner()
    a = 12
    return inner

The 'inner' function closes over the outer scope. Each function with
cellvars packs them into a heap-allocated structure, the closure scope.

The closure scope is passed into 'inner' when called from within outer.

The execution of 'def' creates a NumbaFunction, which has itself as the
 m_self attribute. So when 'inner' is invoked from Python, the numba
 wrapper function gets called with NumbaFunction object and the args
 tuple. The closure scope is then set in NumbaFunction.func_closure.

The closure scope is an extension type with the cellvars as attributes.
Closure scopes are chained together, since multiple inner scopes may need
to share a single outer scope. E.g.

    def outer(a):
        def inner(b):
            def inner_inner():
                print a, b
            return inner_inner

        return inner(1), inner(2)

We have three live closure scopes here:

    scope_outer = { 'a': a }  # call to 'outer'
    scope_inner_1 = { 'scope_outer': scope_outer, 'b': 1 } # call to 'inner' with b=1
    scope_inner_2 = { 'scope_outer': scope_outer, 'b': 2 } # call to 'inner' with b=2

Function 'inner_inner' defines no new scope, since it contains no cellvars.
But it does contain a freevar from scope_outer and scope_inner, so it gets
scope_inner passed as first argument. scope_inner has a reference to scope
outer, so all variables can be resolved.

These scopes are instances of a numba extension class.
"""
from __future__ import print_function, division, absolute_import
import ast
import ctypes
import logging

import numba.decorators
from numba import *
from numba import error
from numba import visitors
from numba import nodes
from numba import typesystem
from numba import typedefs
from numba import numbawrapper
from numba.exttypes import extension_types
from numba import utils
from numba.type_inference import module_type_inference
from numba.minivect import  minitypes
from numba.symtab import Variable
from numba.exttypes import methodtable
from numba.exttypes import attributetable
logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)

#------------------------------------------------------------------------
# Utilities
#------------------------------------------------------------------------

def is_closure_signature(func_signature):
    return (func_signature is not None and
            func_signature.args and
            func_signature.args[0].is_closure_scope)

#------------------------------------------------------------------------
# Closure Signature Validation (type inference of outer function)
#------------------------------------------------------------------------

# Handle closures during type inference. Mostly performs error checking
# for closure signatures.

def err_decorator(decorator):
    raise error.NumbaError(
            decorator, "Only @jit and @autojit and signature decorators "
                       "are supported")

def check_valid_argtype(argtype_node, argtype):
    if not isinstance(argtype, minitypes.Type):
        raise error.NumbaError(argtype_node, "Invalid type: %r" % (argtype,))

def assert_constant(visit_func, decorator, result_node):
    result = visit_func(result_node)
    if not result.variable.is_constant:
        raise error.NumbaError(decorator, "Expected a constant")

    return result.variable.constant_value

def parse_argtypes(visit_func, decorator, func_def, jit_args):
    argtypes_node = jit_args['argtypes']
    if argtypes_node is None:
        raise error.NumbaError(func_def.args[0],
                               "Expected an argument type")

    argtypes = assert_constant(visit_func, decorator, argtypes_node)

    if not isinstance(argtypes, (list, tuple)):
        raise error.NumbaError(argtypes_node,
                               'Invalid argument for argtypes')
    for argtype in argtypes:
        check_valid_argtype(argtypes_node, argtype)

    return argtypes

def parse_restype(visit_func, decorator, jit_args):
    restype_node = jit_args['restype']

    if restype_node is not None:
        restype = assert_constant(visit_func, decorator, restype_node)
        if isinstance(restype, (str, unicode)):
            signature = utils.process_signature(restype)
            restype = signature.return_type
            argtypes = signature.args
            check_valid_argtype(restype_node, restype)
            for argtype in argtypes:
                check_valid_argtype(restype_node, argtype)
            restype = restype(*argtypes)
        else:
            check_valid_argtype(restype_node, restype)
    else:
        raise error.NumbaError(restype_node, "Return type expected")

    return restype

def handle_jit_decorator(visit_func, func_def, decorator):
    jit_args = module_type_inference.parse_args(
            decorator, ['restype', 'argtypes', 'backend',
                        'target', 'nopython'])

    if decorator.args or decorator.keywords:
        restype = parse_restype(visit_func, decorator, jit_args)
        if restype is not None and restype.is_function:
            signature = restype
        else:
            argtypes = parse_argtypes(visit_func, decorator, func_def, jit_args)
            signature = minitypes.FunctionType(restype, argtypes,
                                               name=func_def.name)
    else: #elif func_def.args:
        raise error.NumbaError(decorator,
                               "The argument types and return type "
                               "need to be specified")
    #else:
    #    signature = minitypes.FunctionType(None, [])

    # TODO: Analyse closure at call or outer function return time to
    # TODO:     infer return type
    # TODO: parse out nopython argument
    return signature

def check_signature_decorator(visit_func, decorator):
    dec = visit_func(decorator)
    type = dec.variable.type
    if type.is_cast and type.dst_type.is_function:
        return type.dst_type
    else:
        err_decorator(decorator)

def process_decorators(env, visit_func, node):
    if not node.decorator_list:
        func_env = env.translation.get_env(node)
        if func_env:
            return func_env.func_signature

        raise error.NumbaError(
            node, "Closure must be decorated with 'jit' or 'autojit'")

    if len(node.decorator_list) > 1:
        raise error.NumbaError(
                    node, "Only one decorator may be specified for "
                          "closure (@jit/@autojit)")

    decorator, = node.decorator_list

    if isinstance(decorator, ast.Name):
        decorator_name = decorator.id
    elif (not isinstance(decorator, ast.Call) or not
              isinstance(decorator.func, ast.Name)):
        err_decorator(decorator)
    else:
        decorator_name = decorator.func.id

    if decorator_name not in ('jit', 'autojit'):
        signature = check_signature_decorator(visit_func, decorator)
    else:
        if decorator_name == 'autojit':
            raise error.NumbaError(
                decorator, "Dynamic closures not yet supported, use @jit")

        signature = handle_jit_decorator(visit_func, node, decorator)

    del node.decorator_list[:]

    if len(signature.args) != len(node.args.args):
        raise error.NumbaError(
            decorator,
            "Expected %d arguments type(s), got %d" % (
                            len(signature.args), len(node.args.args)))

    return signature

#------------------------------------------------------------------------
# Closure Type Inference
#------------------------------------------------------------------------

def outer_scope_field(scope_type):
    return scope_type.attribute_table.to_struct().fields[0]

def lookup_scope_attribute(cur_scope, var_name, ctx=None):
    """
    Look up a variable in the closure scope
    """
    ctx = ctx or ast.Load()
    scope_type = cur_scope.type
    outer_scope_name, outer_scope_type = outer_scope_field(scope_type)


    if var_name in scope_type.attributedict:
        return nodes.ExtTypeAttribute.from_known_attribute(
            value=cur_scope, attr=var_name, ctx=ctx, ext_type=scope_type)
    elif outer_scope_type.is_closure_scope:
        scope = nodes.ExtTypeAttribute.from_known_attribute(
            value=cur_scope, attr=outer_scope_name, ctx=ctx, ext_type=scope_type)

        try:
            return lookup_scope_attribute(scope, var_name, ctx)
        except error.InternalError as e:
            # Re-raise with full scope type
            pass

    # This indicates a bug
    raise error.InternalError(
            scope_type, "Unable to look up attribute", var_name)

CLOSURE_SCOPE_ARG_NAME = '__numba_closure_scope'

class ClosureTransformer(visitors.NumbaTransformer):

    @property
    def outer_scope(self):
        outer_scope = None
        if CLOSURE_SCOPE_ARG_NAME in self.symtab:
            outer_scope = ast.Name(id=CLOSURE_SCOPE_ARG_NAME, ctx=ast.Load())
            outer_scope.variable = self.symtab[CLOSURE_SCOPE_ARG_NAME]
            outer_scope.type = outer_scope.variable.type

        return outer_scope

class ClosureTypeInferer(ClosureTransformer):
    """
    Runs just after type inference after the outer variables types are
    resolved.

    1) run type inferencer on inner functions
    2) build scope extension types pre-order
    3) generate nodes to instantiate scope extension type at call time
    """

    def __init__(self, *args, **kwargs):
        super(ClosureTypeInferer, self).__init__(*args, **kwargs)
        self.warn = kwargs["warn"]

    def visit_FunctionDef(self, node):
        if node.closure_scope is None:
            # Process inner functions and determine cellvars and freevars
            # codes = [c for c in self.constants
            #                if isinstance(c, types.CodeType)]
            process_closures(self.env, node, self.symtab,
                             func_globals=self.func_globals,
                             closures=self.closures,
                             warn=self.warn)

        # cellvars are the variables we own
        cellvars = dict((name, var) for name, var in self.symtab.iteritems()
                                        if var.is_cellvar)
        node.cellvars = cellvars

        logger.debug("Cellvars in function %s: %s", node.name, cellvars)

        outer_scope = self.outer_scope
        if outer_scope:
            outer_scope_type = outer_scope.type
        else:
            outer_scope_type = None

        if not cellvars:
            # No cellvars, so use parent closure scope if this is a closure
            if outer_scope:
                self.update_closures(node, outer_scope_type, None)

            return self.visit_func_children(node)

        # Create closure scope extension type
        cellvar_fields = [(name, var.type)
                              for name, var in cellvars.iteritems()]
        fields = numba.struct(cellvar_fields).fields

        if outer_scope:
            fields.insert(0, ('__numba_base_scope', outer_scope_type))

        class py_class(object):
            pass

        func_name = self.func_name
        py_class.__name__ = '%s_scope' % func_name
        scope_type = typesystem.ClosureScopeType(py_class, outer_scope_type)
        scope_type.unmangled_symtab = dict(fields)

        AttrTable = attributetable.AttributeTable
        scope_type.attribute_table = AttrTable.from_list(py_class=None,
                                                         attributes=fields)

        ext_type = extension_types.create_new_extension_type(
                type, func_name , (object,), {}, scope_type, None)

        # Instantiate closure scope
        logger.debug("Generate closure %s %s %s", node.name, scope_type,
                     outer_scope)

        cellvar_scope = nodes.InstantiateClosureScope(
                    node, ext_type, scope_type, outer_scope)
        node.body.insert(0, cellvar_scope)

        self.update_closures(node, scope_type, ext_type)
        return self.visit_func_children(node)

    def update_closures(self, func_def, scope_type, ext_type):
        """
        Patch closures to get the closure scope as the first argument.
        """
        for closure in func_def.closures:
            # closure.scope_type = scope_type
            closure.func_def.scope_type = scope_type
            closure.ext_type = ext_type

            # patch function parameters
            param = ast.Name(id=CLOSURE_SCOPE_ARG_NAME, ctx=ast.Param())
            param.variable = Variable(scope_type, is_local=True)
            param.type = param.variable.type

            closure.symtab[CLOSURE_SCOPE_ARG_NAME] = param.variable
            closure.func_def.args.args.insert(0, param)
            closure.need_closure_scope = True

            # patch closure signature
            closure.type.add_scope_arg(scope_type)
            closure.func_env.func_signature = closure.type.signature


def get_locals(symtab):
    return dict((name, var) for name, var in symtab.iteritems()
                    if var.is_local)

def process_closures(env, outer_func_def, outer_symtab, **kwds):
    """
    Process closures recursively and for each variable in each function
    determine whether it is a freevar, a cellvar, a local or otherwise.
    """
    import numba.pipeline

    outer_symtab = get_locals(outer_symtab)

    # closure_scope is set on the FunctionDef by TypeInferer
    if outer_func_def.closure_scope is not None:
        closure_scope = dict(outer_func_def.closure_scope, **outer_symtab)
    else:
        closure_scope = outer_symtab

    for closure in outer_func_def.closures:
        logger.debug("process closures: %s %s", outer_func_def.name,
                     closure.func_def.name)
        closure_py_func = None # closure.py_func
        func_env, _ = numba.pipeline.run_pipeline2(
                env,
                closure_py_func,
                closure.func_def,
                closure.type.signature,
                closure_scope=closure_scope,
                function_globals=env.translation.crnt.function_globals,
                pipeline_name='type_infer',
                is_closure=True,
                **kwds)

        closure.func_env = func_env
        closure.symtab = func_env.symtab

        env.translation.push_env(func_env)
        process_closures(env, closure.func_def, func_env.symtab, **kwds)
        env.translation.pop()

#------------------------------------------------------------------------
# Closure Lowering
#------------------------------------------------------------------------

class ClosureSpecializer(ClosureTransformer):
    """
    Lowering of closure instantiation and calling.

        - Instantiates the closure scope and makes the necessary assignments
        - Rewrites local variable accesses to accesses on the instantiated scope
        - Instantiate function with closure scope

    Also rewrite calls to closures.
    """

    def __init__(self, *args, **kwargs):
        super(ClosureSpecializer, self).__init__(*args, **kwargs)
        if not self.ast.cellvars:
            self.ast.cur_scope = self.outer_scope

    def _load_name(self, var_name, is_cellvar=False):
        src = ast.Name(var_name, ast.Load())
        src.variable = Variable.from_variable(self.symtab[var_name])
        src.variable.is_cellvar = is_cellvar
        src.type = src.variable.type
        return src

    def visit_InstantiateClosureScope(self, node):
        """
        Instantiate a closure scope.

        After instantiation, assign the parent scope and all function
        arguments that belong in the scope to the scope.
        """
        ctor = nodes.objconst(node.ext_type.__new__)
        ext_type_arg = nodes.objconst(node.ext_type)
        create_scope = nodes.ObjectCallNode(
                    signature=node.scope_type(object_), func=ctor,
                    args=[ext_type_arg])

        create_scope = create_scope.cloneable
        scope = create_scope.clone
        stats = [create_scope]

        # Chain outer scope - if present - to current scope
        outer_scope = self.outer_scope
        if outer_scope:
            outer_scope_name, outer_scope_type = outer_scope_field(scope.type)
            dst = lookup_scope_attribute(scope, outer_scope_name,
                                         ctx=ast.Store())
            assmt = ast.Assign(targets=[dst], value=outer_scope)
            stats.append(assmt)

        # Assign function arguments that are cellvars
        for arg in self.ast.args.args:
            name = arg.id

            if name in node.scope_type.unmangled_symtab:
                dst = lookup_scope_attribute(scope, name, ast.Store())
                src = self._load_name(name)
                src.variable.assign_in_closure_scope = True
                assmt = ast.Assign(targets=[dst], value=src)
                stats.append(assmt)

        logger.debug("instantiating %s", scope.type)
        self.ast.cur_scope = scope
        return self.visit(nodes.ExpressionNode(stmts=stats, expr=scope))

    def get_qname(self, closure_node):
        ns = '.'.join([self.module_name, self.func_name])
        closure_name = closure_node.name
        qname = "%s.__closure__.%s" % (ns, closure_name)
        return qname

    def visit_ClosureNode(self, node):
        """
        Compile the inner function.
        """
        # Compile closure, skip CFA and type inference
        node.func_env.qualified_name = self.get_qname(node)
        numba.pipeline.run_env(self.env, node.func_env,
                               pipeline_name='compile')

        translator = node.func_env.translator
        # translator.link()
        node.lfunc = translator.lfunc
        node.lfunc_pointer = translator.lfunc_pointer

        if node.need_numba_func:
            return self.create_numba_function(node, node.func_env)
        else:
            func_name = node.func_def.name
            self.symtab[func_name] = Variable(name=func_name, type=node.type,
                                              is_local=True)
            # return nodes.LLVMValueRefNode(node.type, node.lfunc)
            # TODO: Remove assignment altogether!
            # return nodes.NoneNode()
            return nodes.ObjectInjectNode(None, type=object_)

    def create_numba_function(self, node, func_env):
        from numba.codegen import llvmwrapper

        closure_scope = self.ast.cur_scope

        if closure_scope is None:
            closure_scope = nodes.NULL
            scope_type = void.pointer()
        else:
            assert node.func_def.args.args[0].variable.type
            scope_type = closure_scope.type

        self.env.translation.push_env(func_env)
        try:
            node.wrapper_func, node.wrapper_lfunc, methoddef = (
                        llvmwrapper.build_wrapper_function(self.env))
        finally:
            self.env.translation.pop()

        # Keep methoddef alive
        # assert methoddef in node.py_func.live_objects
        modname = self.module_name
        self.keep_alive(modname)

        # Create function signature with closure scope at runtime
        create_numbafunc_signature = node.type(
            void.pointer(),     # PyMethodDef *ml
            object_,            # PyObject *module
            void.pointer(),     # PyObject *code
            scope_type,         # PyObject *closure
            void.pointer(),     # void *native_func
            object_,            # PyObject *native_signature
            object_,            # PyObject *keep_alive
        )

        # Create function with closure scope at runtime
        create_numbafunc = nodes.ptrfromint(
                        numbawrapper.NumbaFunction_NewEx_pointer,
                        create_numbafunc_signature.pointer())

        methoddef_p = ctypes.cast(ctypes.byref(methoddef),
                                  ctypes.c_void_p).value

        args = [
            nodes.const(methoddef_p, void.pointer()),
            nodes.const(modname, object_),
            nodes.NULL,
            closure_scope,
            nodes.const(node.lfunc_pointer, void.pointer()),
            nodes.const(node.type.signature, object_),
            nodes.NULL, # nodes.const(node.py_func, object_),
        ]

        func_call = nodes.NativeFunctionCallNode(
                            signature=create_numbafunc_signature,
                            function_node=create_numbafunc,
                            args=args)

        result = func_call

        #stats = [nodes.inject_print(nodes.const("calling...", c_string_type)),
        #         result]
        #result = ast.Suite(body=stats)
        result = self.visit(result)
        return result

    def visit_Name(self, node):
        "Resolve cellvars and freevars"
        is_param = isinstance(node.ctx, ast.Param)
        if not is_param and (node.variable.is_cellvar or
                             node.variable.is_freevar):
            logger.debug("Function %s, lookup %s in scope %s: %s",
                          self.ast.name, node.id, self.ast.cur_scope.type,
                          self.ast.cur_scope.type.attribute_table)
            attr = lookup_scope_attribute(self.ast.cur_scope,
                                          var_name=node.id, ctx=node.ctx)
            return self.visit(attr)
        else:
            return node

    def retrieve_closure_from_numbafunc(self, node):
        """
        Retrieve the closure scope from ((NumbaFunctionObject *)
                                             numba_func).func_closure
        """
        # TODO: use llvmwrapper.get_closure_scope()
        pointer = nodes.ptrfromobj(node.func)
        type = typedefs.NumbaFunctionObject.ref()
        closure_obj_struct = nodes.CoercionNode(pointer, type)
        cur_scope = nodes.StructAttribute(closure_obj_struct, 'func_closure',
                                          ctx=ast.Load(), type=type)
        return cur_scope

    def visit_ClosureCallNode(self, node):
        if node.closure_type.closure.need_closure_scope:
            if (self.ast.cur_scope is None or
                    self.ast.cur_scope.type != node.closure_type):
                # Call to closure from outside outer function
                # TODO: optimize calling a closure from an inner function, e.g.
                # def outer():
                #     a = 1
                #     def inner1(): print a
                #     def inner2(): inner1()
                cur_scope = self.retrieve_closure_from_numbafunc(node)
            else:
                # Call to closure from within outer function
                cur_scope = self.ast.cur_scope

            # node.args[0] = cur_scope
            node.args.insert(0, cur_scope)

        self.generic_visit(node)
        return node

    def visit_ClosureScopeLoadNode(self, node):
        return self.ast.cur_scope or nodes.NULL_obj



# -*- coding: utf-8 -*-

"""
Extension type types.
"""

from numba.minivect import minitypes

from numba.traits import traits, Delegate
from numba.typesystem import NumbaType

@traits
class ExtensionType(NumbaType, minitypes.ObjectType):
    """
    Extension type Numba type.

    Available to users through MyExtensionType.exttype (or
    numba.typeof(MyExtensionType).
    """

    is_extension = True
    is_final = False

    methoddict = Delegate('vtab_type')
    untyped_methods = Delegate('vtab_type')
    specialized_methods = Delegate('vtab_type')
    methodnames = Delegate('vtab_type')
    add_method = Delegate('vtab_type')

    attributedict = Delegate('attribute_table')
    attributes = Delegate('attribute_table')

    def __init__(self, py_class, **kwds):
        super(ExtensionType, self).__init__(**kwds)
        assert isinstance(py_class, type), ("Must be a new-style class "
                                            "(inherit from 'object')")
        self.name = py_class.__name__
        self.py_class = py_class
        self.extclass = None

        self.symtab = {}  # attr_name -> attr_type

        self.compute_offsets(py_class)

        self.attribute_table = None
        self.vtab_type = None

        self.parent_attr_struct = None
        self.parent_vtab_type = None
        self.parent_type = getattr(py_class, "__numba_ext_type", None)

    def compute_offsets(self, py_class):
        from numba.exttypes import extension_types

        self.vtab_offset = extension_types.compute_vtab_offset(py_class)
        self.attr_offset = extension_types.compute_attrs_offset(py_class)


# ______________________________________________________________________
# @jit

class JitExtensionType(ExtensionType):
    "Type for @jit extension types"

    is_jit_extension = True

    def __repr__(self):
        return "<JitExtension %s>" % self.name

    def __str__(self):
        if self.attribute_table:
            return "<JitExtension %s(%s)>" % (
                    self.name, self.attribute_table.strtable())
        return repr(self)

# ______________________________________________________________________
# @autojit

class AutojitExtensionType(ExtensionType):
    "Type for @autojit extension types"

    is_autojit_extension = True

    def __repr__(self):
        return "<AutojitExtension %s>" % self.name

    def __str__(self):
        if self.attribute_table:
            return "<AutojitExtension %s(%s)>" % (
                    self.name, self.attribute_table.strtable())
        return repr(self)




# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import os

root = os.path.dirname(os.path.abspath(__file__))

def get_include():
    return os.path.join(root, 'include')



#! /usr/bin/env python
from __future__ import print_function, division, absolute_import
# -*- coding: utf-8 -*-
# ______________________________________________________________________

import ctypes
import ctypes.util

from numba import *

# ______________________________________________________________________

c_void_pp = ctypes.POINTER(ctypes.c_void_p)

def get_libc ():
    return ctypes.CDLL(ctypes.util.find_library('c'))

def get_stdio_streams ():
    '''
    Returns file pointers (FILE *) as Python integers for the C stdio
    stdin, stdout, and stderr streams.
    '''
    ret_val = None
    if hasattr(ctypes.pythonapi, 'stdin'):
        # Linux
        _stdio_files = (ctypes.c_void_p.in_dll(ctypes.pythonapi, sym)
                        for sym in ('stdin', 'stdout', 'stderr'))
        ret_val = tuple(c_void_pp(file_p)[0] for file_p in _stdio_files)
    elif hasattr(ctypes.pythonapi, '__stdinp'):
        # OSX
        _stdio_files = (ctypes.c_void_p.in_dll(ctypes.pythonapi, sym)
                        for sym in ('__stdinp', '__stdoutp', '__stderrp'))
        ret_val = tuple(c_void_pp(file_p)[0] for file_p in _stdio_files)
    else:
        libc = get_libc()
        if hasattr(libc, '__getreent'):
            # Cygwin
            ret_val = tuple(ctypes.cast(libc.__getreent(), c_void_pp)[1:4])
        elif hasattr(libc, '__iob_func'):
            # MSVC
            ret_val = tuple(ctypes.cast(libc.__iob_func(), c_void_pp)[0:3])
        else:
            raise NotImplementedError("Unsupported platform, don't know how to "
                                      "find pointers to stdio streams!")
    return ret_val

def get_stream_as_node(fp):
    return nodes.CoercionNode(nodes.ConstNode(fp, Py_uintptr_t),
                              void.pointer())

# ______________________________________________________________________

def main ():
    _, stdout, _ = get_stdio_streams()
    PyObject_Print = ctypes.pythonapi.PyObject_Print
    PyObject_Print.restype = ctypes.c_int
    PyObject_Print.argtypes = ctypes.py_object, ctypes.c_void_p, ctypes.c_int
    PyObject_Print(get_stdio_streams, stdout, 1)
    PyObject_Print('\n\n', stdout, 1)

# ______________________________________________________________________

if __name__ == "__main__":
    main()

# ______________________________________________________________________
# End of stdio_util.py



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from .external import ExternalLibrary

__all__ = []
all = {}

def _import_all():
    global __all__
    mods = ['pyapi',
            'libc']
    for k in mods:
        mod = __import__(__name__ + '.' + k, fromlist=['__all__'])
        __all__.extend(mod.__all__)
        for k in mod.__all__:
            all[k] = globals()[k] = getattr(mod, k)

_import_all()

def default_external_library(context):
    '''Build an external function library
        
    context --- numba context
    '''
    extlib = ExternalLibrary(context)
    
    for fncls in all.values():
        extlib.add(fncls())
    return extlib



cimport numpy as cnp
import numpy as np

import hashlib

cdef extern from "perfecthash.h":
    void _PyCustomSlots_bucket_argsort(uint16_t *p, uint8_t *binsizes,
                                       uint16_t *number_of_bins_by_size)

def bucket_argsort(cnp.ndarray[uint16_t, mode='c'] p,
                   cnp.ndarray[uint8_t, mode='c'] binsizes,
                   cnp.ndarray[uint16_t, mode='c'] number_of_bins_by_size):
    _PyCustomSlots_bucket_argsort(&p[0], &binsizes[0],
                                  &number_of_bins_by_size[0])

def get_random_hashes(rng, nitems):
    return rng.randint(-2**31, 2**31-1, size=nitems).astype(np.uint64)

def draw_hashes(rng, nitems):
    assert sizeof(long) >= 4

    hashes = get_random_hashes(rng, nitems)
    hashes <<= 32
    hashes |= get_random_hashes(rng, nitems)

    return hashes

def perfect_hash(cnp.ndarray[uint64_t] hashes, int repeat=1):
    """Used for testing. Takes the hashes as input, and returns
       a permutation array and hash parameters:

       (p, r, m_f, m_g, d)
    """
    cdef PyCustomSlots_Table_64_64 table
    table.base.flags = 0
    table.base.n = 64
    table.base.b = 64
    table.base.entries = &table.entries_mem[0]
    for i in range(64):
        table.entries_mem[i].id = hashes[i]
        table.entries_mem[i].ptr = NULL

    cdef int r
    for r in range(repeat):
        PyCustomSlots_PerfectHash(&table.base, &hashes[0])

    d = np.zeros(64, dtype=np.uint16)
    p = np.zeros(64, dtype=np.uint16)

    for i in range(64):
        p[i] = table.entries_mem[i].id & 0xFF
        d[i] = table.d[i]

    return p, table.base.r, table.base.m_f, table.base.m_g, d

#cdef extern from "md5sum.h":
#    ctypedef struct MD5_CTX:
#        uint32_t i[2]
#        uint32_t buf[4]
#        unsigned char in_ "in"[64]
#        unsigned char digest[16]
#
#    void MD5Init(MD5_CTX *mdContext)
#    void MD5Update(MD5_CTX *mdContext, unsigned char *inBuf,
#                   unsigned int inLen)
#    void MD5Final(MD5_CTX *mdContext)
#
#cdef extern from "hash.h":
#    uint64_t hash_crapwow64(unsigned char *buf, uint64_t len, uint64_t seed)
#
#def crapwowbench(int repeat=1):
#    cdef int r
#    cdef MD5_CTX ctx
#    for r in range(repeat):
#        hash_crapwow64("asdf", 4, 0xf123456781234567)
#
#
#def md5bench(int repeat=1):
#    cdef int r
#    cdef MD5_CTX ctx
#    for r in range(repeat):
#        MD5Init(&ctx)
#        MD5Update(&ctx, "asdf", 4)
#        MD5Final(&ctx)




# Adapted from cython/tests/run/for_in_range.pyx

from numba.testing.test_support import *

@autojit_py3doc(warn=False)
def test_modify():
    """
    >>> test_modify()
    0
    1
    2
    3
    4
    <BLANKLINE>
    (4L, 0L)
    """
    n = 5
    for i in range(n):
        print(i)
        n = 0
    print('')
    return i,n

@autojit_py3doc(warn=False)
def test_negindex():
    """
    >>> test_negindex()
    6
    5
    4
    3
    2
    (2L, 0L)
    """
    n = 5
    for i in range(n+1, 1, -1):
        print(i)
        n = 0
    return i,n

@autojit_py3doc(warn=False)
def test_negindex_inferred():
    """
    >>> test_negindex_inferred()
    5
    4
    3
    2
    (2L, 0L)
    """
    n = 5
    for i in range(n, 1, -1):
        print(i)
        n = 0
    return i,n

@autojit_py3doc(warn=False)
def test_fix():
    """
    >>> test_fix()
    0
    1
    2
    3
    4
    <BLANKLINE>
    4L
    """
    for i in range(5):
        print(i)
    print('')
    return i

@autojit_py3doc(warn=False)
def test_break():
    """
    >>> test_break()
    0
    1
    2
    <BLANKLINE>
    (2L, 0L)
    """
    n = 5
    for i in range(n):
        print(i)
        n = 0
        if i == 2:
            break
    else:
        print("FAILED!")
    print('')
    return i, n

@autojit_py3doc
def test_else_clause1():
    """
    >>> test_else_clause1()
    0
    1
    2
    """
    for i in range(10):
        if i > 2:
            break
        print(i)
    else:
        print("else clause")

@autojit_py3doc
def test_else_clause2():
    """
    >>> test_else_clause2()
    0
    1
    2
    else clause
    """
    for i in range(10):
        if i > 2:
            continue
        print(i)
    else:
        print("else clause")

@autojit_py3doc
def test_else_clause3():
    """
    >>> test_else_clause3()
    0
    1
    2
    else clause
    """
    for i in range(3):
        if i > 2 and i < 2:
            continue
        print(i)
    else:
        print("else clause")

@autojit_py3doc(warn=False)
def test_else_clause4():
    """
    >>> test_else_clause4()
    inner 0
    i 0
    else clause 1 0 9
    i 1
    else clause 2 0 9
    i 2
    else clause 3 0 9
    i 3
    else clause 4 0 9
    i 4
    else clause 5 0 9
    i 5
    else clause 6 0 9
    i 6
    else clause 7 0 9
    i 7
    else clause 8 0 9
    i 8
    else clause 9 0 9
    i 9
    else clause
    """
    for i in range(10):
        for j in range(10):
            for k in range(10):
                if i == j and j == k:
                    print("inner " + str(i))
                    break
                else:
                    continue
            else:
                print("else clause " + str(i) + ' ' + str(j) + ' ' + str(k))
            break
        else:
            print("else clause " + str(i) + ' ' + str(j))

        print("i " + str(i))
    else:
        print("else clause")

@autojit_py3doc
def test_return():
    """
    >>> test_return()
    0
    1
    2
    (2L, 0L)
    """
    n = 5
    for i in range(n):
        print(i)
        n = 0
        if i == 2:
            return i,n
    print('')
    return "FAILED!"

@autojit_py3doc
def test_return2():
    """
    >>> test_return2()
    0
    1
    2
    2L
    """
    n = 5
    for i in range(n):
        print(i)
        n = 0
        for j in range(n):
            return 0
        else:
            if i < 2:
                continue
            elif i == 2:
                for j in range(i):
                    return i
                print("FAILED!")
            print("FAILED!")
        print("FAILED!")
    return -1

#print test_negindex()
#test_else_clause2()
#test_else_clause3()
#test_else_clause4()
#test_return2()
testmod()



# -*- coding: utf-8 -*-

"""
Virtual method lookup written in Numba.
"""

from __future__ import division, absolute_import

import ctypes.util

import numba
from numba import *
from numba.exttypes import virtual


table_t = virtual.PyCustomSlots_Table

table_t_pp = table_t.pointer().pointer()
char_p = char.pointer()
void_p = void.pointer()
uint16_p = uint16.pointer()

# This is a bad idea!
# displacements_offset = table_t.offsetof('d')
displacements_offset = ctypes.sizeof(table_t.to_ctypes())

#------------------------------------------------------------------------
# Some libc functions
#------------------------------------------------------------------------

libc = ctypes.CDLL(ctypes.util.find_library('c'))

abort = libc.abort
abort.restype = None
abort.argtypes = []

printf = libc.printf
printf.restype = ctypes.c_int
printf.argtypes = [ctypes.c_char_p]

@jit(void_p(table_t_pp, uint64), wrap=False, nopython=True)
def lookup_method(table_pp, prehash):
    """
    Look up a method in a PyCustomSlots_Table ** given a prehash.

        PyCustomSlots *vtab = atomic_load(table_pp)

        f = (prehash >> vtab->r) & vtab->m_f
        g = vtab->d[prehash & vtab->m_g]
        PyCustomSlot_Entry *entry = vtab->entries[f ^ g]

        if (entry->id == prehash) {
            void *vmethod = entry.ptr
            call vmethod(obj, ...)
        } else {
            PyObject_Call(obj, "method", ...)
        }

    Note how the object stores a vtable **, instead of a vtable *. This
    indirection allows producers of the table to generate a new table
    (e.g. after adding a specialization) and replace the table for all
    live objects.

    We then atomically load the table, to allow using the table without
    the GIL (and having the GIL holding thread update the table and rewrite
    the pointer).

    Hence the table pointer should *not* be cached by callers, since the
    first table miss can trigger a compilation you will want to find the
    second time around.
    """
    table_p = table_pp[0]
    table = table_p[0]
    displacements = uint16_p(char_p(table_p) + displacements_offset)

    # Compute f
    f = (prehash >> table.r) & table.m_f

    # Compute g
    # g = table.d[prehash & table.m_g]
    g = displacements[prehash & table.m_g]

    entry = table.entries[f ^ g]

    if entry.id == prehash:
        return entry.ptr
    else:
        return numba.NULL

@jit(void_p(table_t_pp, uint64, char_p), wrap=False, nopython=True)
def lookup_and_assert_method(table_pp, prehash, method_name):
    result = lookup_method(table_pp, prehash)
    if result == numba.NULL:
        # printf("Error: expected method %s to be available\n", method_name)
        # print "Error: expected method", method_name,  "to be available"
        printf("NumbaError: expected method ")
        printf(method_name)
        printf(" to be available\n")
        abort()

    return result



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import llvm.core

def declare(numba_cdef, env, global_module):
    """
    Declare a NumbaCDefinition in the current translation environment.
    """
    # print numba_cdef
    specialized_cdef = numba_cdef(env, global_module)
    lfunc = specialized_cdef.define(global_module) #, optimize=False)
    assert lfunc.module is global_module
    return specialized_cdef, lfunc

registered_utilities = []

def register(utility):
    registered_utilities.append(utility)
    return utility

def load_utilities():
    from . import library
    from . import numbacdef
    from . import refcounting

class CBuilderLibrary(object):
    """
    Library of cbuilder functions.
    """

    def __init__(self):
        self.module = llvm.core.Module.new("cbuilderlib")
        self.funcs = {}

    def declare_registered(self, env):
        "Declare all utilities in our module"
        load_utilities()
        for registered_utility in registered_utilities:
            self.declare(registered_utility, env, self.module)

    def declare(self, numba_cdef, env, llvm_module):
        if numba_cdef not in self.funcs:
            specialized_cdef, lfunc = declare(numba_cdef, env, self.module)
            self.funcs[numba_cdef] = specialized_cdef, lfunc
        else:
            specialized_cdef, lfunc = self.funcs[numba_cdef]

        name = numba_cdef._name_
        lfunc_type = specialized_cdef.signature()
        lfunc = llvm_module.get_or_insert_function(lfunc_type, name)

        return lfunc

    def link(self, llvm_module):
        """
        Link the CBuilder library into the target module.
        """
        llvm_module.link_in(self.module, preserve=True)


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from numba.nodes import *
import numba.nodes

class CoercionNode(ExprNode):
    """
    Coerce a node to a different type
    """

    _fields = ['node']
    _attributes = ['type', 'name']

    def __new__(cls, node, dst_type, name=''):
        if isinstance(node, CoercionNode) and node.type == dst_type:
            return node

        return super(CoercionNode, cls).__new__(cls, node, dst_type, name=name)

    def __init__(self, node, dst_type, name=''):
        if node is self:
            # We are trying to coerce a CoercionNode which already has the
            # right type, so __new__ returns a CoercionNode, which then results
            # in __init__ being called
            return

        type = getattr(node, 'type', None) or node.variable.type
        if dst_type.is_pointer and type.is_int:
            assert type == Py_uintptr_t, type

        self.type = dst_type
        self.variable = Variable(dst_type)
        self.name = name

        self.node = self.verify_conversion(dst_type, node)

        if (dst_type.is_object and not node.variable.type.is_object and
                isinstance(node, numba.nodes.ArrayAttributeNode)):
            self.node = self.coerce_numpy_attribute(node)

    def coerce_numpy_attribute(self, node):
        """
        Numpy array attributes, such as 'data', get rewritten to direct
        accesses. Since they are being coerced back to objects, use a generic
        attribute access instead.
        """
        node = ast.Attribute(value=node.array, attr=node.attr_name,
                             ctx=ast.Load())
        node.variable = Variable(object_)
        node.type = object_
        return node

    @property
    def dst_type(self):
        """
        dst_type is always the same as type, and 'type' is kept consistent
        with Variable.type
        """
        return self.type

    @classmethod
    def coerce(cls, node_or_nodes, dst_type):
        if isinstance(node_or_nodes, list):
            return [cls(node, dst_type) for node in node_or_nodes]
        return cls(node_or_nodes, dst_type)

    def verify_conversion(self, dst_type, node):
        if ((node.variable.type.is_complex or dst_type.is_complex) and
            (node.variable.type.is_object or dst_type.is_object)):
            if dst_type.is_complex:
                complex_type = dst_type
            else:
                complex_type = node.variable.type

            if not complex_type == complex128:
                node = CoercionNode(node, complex128)

        return node

    def __repr__(self):
        return "Coerce(%s, %s)" % (self.type, self.node)

class CastNode(ExprNode):
    """
    Explicit cast by user, e.g. double(value)
    """

    _fields = ["arg"]

    def __init__(self, node, type):
        self.arg = node
        self.type = type


class PromotionNode(ExprNode):
    """
    Coerces a variable of some type to another type for a phi node in a
    successor block.
    """

    _fields = ['node']

    def __init__(self, **kwargs):
        super(PromotionNode, self).__init__(**kwargs)
        self.variable = self.node.variable

class CoerceToObject(CoercionNode):
    "Coerce native values to objects"

class CoerceToNative(CoercionNode):
    "Coerce objects to native values"


class DeferredCoercionNode(ExprNode):
    """
    Coerce to the type of the given variable. The type of the variable may
    change in the meantime (e.g. may be promoted or demoted).
    """

    _fields = ['node']

    def __init__(self, node, variable):
        self.node = node
        self.variable = variable

class UntypedCoercion(ExprNode):
    """
    Coerce a node to the destination type. The node need not yet have a
    type or variable.
    """

    _fields = ['node']

    def __init__(self, node, type):
        self.node = node
        self.type = type



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import numpy as np
import numba as nb
from numba import jit, int_, void, float64

@jit
class Foo(object):
    @void(int_)
    def __init__(self, size):
        self.arr = np.zeros(size, dtype=float)
        self.type = nb.typeof(self.arr)

assert Foo(10).type == float64[:]



import llvm.core

class LLVMConstantsManager(object):
    """
    Manage global constants. The result should be linked into the consumer
    LLVM module.
    """

    def __init__(self):
        self.module = llvm.core.Module.new("numba_constants")

        # py_constant -> llvm value
        self.constant_values = {}

    def link(self, dst_module):
        dst_module.link_in(self.module, preserve=True)

    def get_string_constant(self, const_str):
        if const_str in self.constant_values:
            ret_val = self.constant_values[const_str]
        else:
            lconst_str = llvm.core.Constant.stringz(const_str)
            ret_val = self.module.add_global_variable(
                lconst_str.type, "__STR_%d" % (len(self.constant_values),))

            ret_val.linkage = llvm.core.LINKAGE_LINKONCE_ODR
            ret_val.initializer = lconst_str
            ret_val.is_global_constant = True

            self.constant_values[const_str] = ret_val

        return ret_val



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import collections

from numba import typesystem
from numba.nodes import *
import numba.nodes
from numba.ndarray_helpers import PyArrayAccessor

#----------------------------------------------------------------------------
# External Utilities
#----------------------------------------------------------------------------

def is_constant_index(node):
    return (isinstance(node, ast.Index) and
            isinstance(node.value, ConstNode))

def is_newaxis(node):
    v = node.variable
    return (is_constant_index(node) and
            node.value.pyval is None) or v.type.is_newaxis or v.type.is_none

def is_ellipsis(node):
    return is_constant_index(node) and node.value.pyval is Ellipsis


#----------------------------------------------------------------------------
# Internal Utilities
#----------------------------------------------------------------------------

def _const_int(X):
    return llvm.core.Constant.int(llvm.core.Type.int(), X)

def get_shape(builder, tbaa, shape_pointer, ndim):
    "Load the shape values from an ndarray"
    shape_metadata = tbaa.get_metadata(typesystem.numpy_shape)

    for i in range(ndim):
        shape_ptr = builder.gep(shape_pointer, [_const_int(i)])
        extent = builder.load(shape_ptr)
        extent.set_metadata("tbaa", shape_metadata)
        yield extent

def get_strides(builder, tbaa, strides_pointer, ndim):
    "Load the stride values from an ndarray"
    stride_metadata = tbaa.get_metadata(typesystem.numpy_strides)

    for i in range(ndim):
        stride_ptr = builder.gep(strides_pointer, [_const_int(i)])
        stride = builder.load(stride_ptr)
        stride.set_metadata("tbaa", stride_metadata)
        yield stride

#----------------------------------------------------------------------------
# Utilities for ndarray attribute preloading
#----------------------------------------------------------------------------

# NOTE: See also optimize.py:Preloader

def make_shape_phis(context, builder, attr, name, ndim):
    "Make phi values for the extents of an array variable merge"
    ltype = npy_intp.to_llvm(context)
    for i in range(ndim):
        yield builder.phi(ltype, "%s%d(%s)" % (attr, i, name))

def make_strides_phis(context, builder, name, ndim):
    "Make phi values for the strides of an array variable merge"
    ltype = typesystem.numpy_strides.to_llvm(context)
    for i in range(ndim):
        yield builder.phi(ltype, "strides%d(%s)" % name)

def make_preload_phi(context, builder, phi_node):
    """
    Build phi values for preloaded data/shape/stride values.
    """
    var = phi_node.variable
    name = var.unmangled_name
    ndim = var.type.ndim

    if var.preload_data:
        ltype = char.pointer().to_llvm(context)
        var.preloaded_data = builder.phi(ltype,
                                         "data(%s)" % name)
    if var.preload_shape:
        var.preloaded_shape = tuple(make_shape_phis(context, builder,
                                                    "shape", name, ndim))
    if var.preload_strides:
        var.preloaded_strides = tuple(make_shape_phis(context, builder,
                                                      "strides", name, ndim))

def update_preloaded_phi(phi_var, incoming_var, llvm_incoming_block):
    if phi_var.preload_data:
        phi_var.preloaded_data.add_incoming(incoming_var.preloaded_data,
                                            llvm_incoming_block)

    if phi_var.preload_shape:
        for extent, incoming in zip(phi_var.preloaded_shape,
                                    incoming_var.preloaded_shape):
            extent.add_incoming(incoming, llvm_incoming_block)

    if phi_var.preload_strides:
        for stride, incoming in zip(phi_var.preloaded_strides,
                                    incoming_var.preloaded_strides):
            stride.add_incoming(incoming, llvm_incoming_block)


#----------------------------------------------------------------------------
# NumPy Array Attributes
#----------------------------------------------------------------------------

class DataPointerNode(ExprNode):

    _fields = ['node', 'slice']

    def __init__(self, node, slice, ctx):
        self.node = node
        self.slice = slice
        self.type = node.type.dtype
        self.variable = Variable(self.type)
        self.ctx = ctx

    def data_descriptors(self, builder, tbaa, llvm_value):
        '''
        Returns a tuple of (dptr, strides)
        - dptr:    a pointer of the data buffer
        - strides: a pointer to an array of stride information;
                   has `ndim` elements.
        '''
        acc = PyArrayAccessor(builder, llvm_value, tbaa, self.type)

        var = self.node.variable

        # Load the data pointer. Use preloaded value if available
        if var.preload_data:
            # print "using preloaded data", var.preloaded_data
            dptr = var.preloaded_data
        else:
            dptr = acc.data

        # Load the strides. Use preloaded values if available
        if var.preload_strides:
            # print "using preloaded strides", var.preloaded_strides
            strides = var.preloaded_strides
        else:
            strides_pointer = acc.strides
            strides = get_strides(builder, tbaa, strides_pointer, self.ndim)

        assert dptr is not None
        assert strides is not None

        return dptr, strides

    def subscript(self, translator, tbaa, llvm_value, indices):
        builder = translator.builder
        caster = translator.caster
        context = translator.context

        offset = _const_int(0)

        if not isinstance(indices, collections.Iterable):
            indices = (indices,)

        dptr, strides = self.data_descriptors(builder, tbaa, llvm_value)

        for i, (stride, index) in enumerate(zip(strides, indices)):
            index = caster.cast(index, stride.type, unsigned=False)
            offset = caster.cast(offset, stride.type, unsigned=False)
            offset = builder.add(offset, builder.mul(index, stride))

        data_ty = self.type.to_llvm(context)
        data_ptr_ty = llvm.core.Type.pointer(data_ty)

        dptr_plus_offset = builder.gep(dptr, [offset])

        ptr = builder.bitcast(dptr_plus_offset, data_ptr_ty)
        return ptr

    @property
    def ndim(self):
        return self.node.type.ndim

    def __repr__(self):
        return "%s.data" % self.node

class ArrayAttributeNode(ExprNode):
    is_read_only = True

    _fields = ['array']

    def __init__(self, attribute_name, array):
        self.array = array
        self.attr_name = attribute_name

        self.array_type = array.variable.type
        if attribute_name == 'ndim':
            type = minitypes.int_
        elif attribute_name in ('shape', 'strides'):
            type = typesystem.SizedPointerType(typesystem.intp,
                                                size=self.array_type.ndim)
        elif attribute_name == 'data':
            type = self.array_type.dtype.pointer()
        else:
            raise error._UnknownAttribute(attribute_name)

        self.type = type

    def __repr__(self):
        return "%s.%s" % (self.array, self.attr_name)

class ShapeAttributeNode(ArrayAttributeNode):
    # NOTE: better do this at code generation time, and not depend on
    #       variable.lvalue
    _fields = ['array']

    def __init__(self, array):
        super(ShapeAttributeNode, self).__init__('shape', array)
        self.array = array
        self.element_type = typesystem.intp
        self.type = minitypes.CArrayType(self.element_type,
                                         array.variable.type.ndim)

#----------------------------------------------------------------------------
# NumPy Array Creation
#----------------------------------------------------------------------------

class ArrayNewNode(ExprNode):
    """
    Allocate a new array given the attributes.
    """

    _fields = ['data', 'shape', 'strides', 'base']

    def __init__(self, type, data, shape, strides, base=None, **kwargs):
        super(ArrayNewNode, self).__init__(**kwargs)
        self.type = type
        self.data = data
        self.shape = shape
        self.strides = strides
        self.base = base

class ArrayNewEmptyNode(ExprNode):
    """
    Allocate a new array with data.
    """

    _fields = ['shape']

    def __init__(self, type, shape, is_fortran=False, **kwargs):
        super(ArrayNewEmptyNode, self).__init__(**kwargs)
        self.type = type
        self.shape = shape
        self.is_fortran = is_fortran


#----------------------------------------------------------------------------
# Nodes for NumPy calls
#----------------------------------------------------------------------------

shape_type = npy_intp.pointer()
void_p = void.pointer()

class MultiArrayAPINode(NativeCallNode):

    def __init__(self, name, signature, args):
        super(MultiArrayAPINode, self).__init__(signature, args,
                                                llvm_func=None)
        self.func_name = name

def PyArray_NewFromDescr(args):
    """
    Low-level specialized equivalent of ArrayNewNode
    """
    signature = object_(
        object_,    # subtype
        object_,    # descr
        int_,       # ndim
        shape_type, # shape
        shape_type, # strides
        void_p,     # data
        int_,       # flags
        object_,    # obj
    )

    return MultiArrayAPINode('PyArray_NewFromDescr', signature, args)

def PyArray_SetBaseObject(args):
    signature = int_(object_, object_)
    return MultiArrayAPINode('PyArray_SetBaseObject', signature, args)

def PyArray_UpdateFlags(args):
    return MultiArrayAPINode('PyArray_UpdateFlags', void(object_, int_), args)

def PyArray_Empty(args, name='PyArray_Empty'):
    nd, shape, dtype, fortran = args
    return_type = minitypes.ArrayType(dtype, nd)
    signature = return_type(
                int_,                   # nd
                npy_intp.pointer(),     # shape
                object_,                # dtype
                int_)                   # fortran
    return MultiArrayAPINode(name, signature, args)

def PyArray_Zeros(args):
    return PyArray_Empty(args, name='PyArray_Zeros')



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import numpy as np

import numba
from numba.minivect import minitypes
from numba import typesystem
from numba.type_inference.module_type_inference import register, register_inferer


@register(numba)
def typeof(expr_type):
    from numba import nodes

    type = typesystem.CastType(expr_type)
    return nodes.const(expr_type, type)



# -*- coding: utf-8 -*-
"""
Hold metadata for instructions.
"""
from __future__ import print_function, division, absolute_import

import llvm.core

from numba import *

def _typename(type):
    typename = str(type)
    typename = typename.replace("float", "float32")
    typename = typename.replace("double", "float64")
    typename = typename.replace("long double", "float128")
    return typename

def typename(type):
    "Get the TBAA type name"
    if type.is_tbaa:
        return type.name
    else:
        return _typename(type)

def is_tbaa_type(type):
    return (type.is_pointer or type.is_tbaa or
            type.is_object or type.is_array)

class TBAAMetadata(object):
    """
    Type Based Alias Analysis metadata. This defines a type tree where
    types in different branches are known not to alias each other. Only
    the ancestors of a type node alias that node.
    """

    def __init__(self, module):
        self.module = module
        self.metadata_cache = {}
        self.initialize()
        self.unique_number = 0

    def initialize(self):
        self.root = self.make_metadata("root", root=None)
        self.char_pointer = self.make_metadata("char *", root=self.root)
        self.metadata_cache[char.pointer()] = self.char_pointer

    def make_metadata(self, typename, root, is_constant=False):
        operands = [self.get_string(typename)]
        if root is not None:
            assert isinstance(root, llvm.core.MetaData)
            operands.append(root)

            if is_constant:
                constant = llvm.core.Constant.int(llvm.core.Type.int(64), 1)
                operands.append(constant)

        node = llvm.core.MetaData.get(self.module, operands)
        llvm.core.MetaData.add_named_operand(self.module, "tbaa", node)

        # print "made metadata", self.module.id, typename # , node, root
        return node

    def make_unique_metadata(self, root, is_constant=False):
        result = self.make_metadata("unique%d" % self.unique_number, root,
                                    is_constant)
        self.unique_number += 1
        return result

    def get_string(self, string):
        return llvm.core.MetaDataString.get(self.module, string)

    def find_root(self, type):
        """
        Find the metadata root of a type. E.g. if we have tbaa(sometype).pointer(),
        we want to return tbaa(sometype).root.pointer()
        """
        # Find TBAA base type
        n = 0
        while type.is_pointer:
            type = type.base_type
            n += 1

        if type.is_tbaa:
            # Return TBAA base type pointer composition
            root_type = type.root
            for i in range(n):
                root_type = root_type.pointer()

            # Define root type metadata
            root = self.get_metadata(root_type)
        else:
            # Not a TBAA root type, alias anything
            root = self.char_pointer

        return root

    def get_metadata(self, type, typeroot=None):
        if not is_tbaa_type(type):
            return None

        if type in self.metadata_cache:
            return self.metadata_cache[type]

        # Build metadata
        if typeroot:
            root = self.metadata_cache[typeroot]
        elif type.is_tbaa:
            if type.root in self.metadata_cache:
                root = self.metadata_cache[type.root]
            else:
                root = self.get_metadata(type.root)
        else:
            root = self.find_root(type)

        node = self.make_metadata(typename(type), root,
                                  is_constant="const" in type.qualifiers)

        # Cache result
        self.metadata_cache[type] = node

        return node

    def set_metadata(self, load_instr, metadata):
        load_instr.set_metadata("tbaa", metadata)

    def set_tbaa(self, load_instr, type):
        md = self.get_metadata(type)
        if md is not None:
            self.set_metadata(load_instr, md)



"""
>>> range_ret1()
[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
>>> range_ret2()
[1, 2, 3, 4]
>>> range_ret3()
[10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4]

>>> forward1()
0 1 2 3 4 5 6 7 8 9 done
>>> forward2()
1 2 3 4 done
>>> forward3()
5 8 11 14 done

>>> backward1()
10 7 4 done
>>> backward2()
done
>>> backward3()
-5 -8 -11 -14 done

>>> empty_assign()
14
>>> last_value()
Warning 92:10: local variable 'i' might be referenced before assignment
9
"""
from __future__ import print_function
from numba import *

@autojit
def range_ret1():
    return range(10)

@autojit
def range_ret2():
    return range(1, 5)

@autojit
def range_ret3():
    return range(10, -5, -1)

@autojit
def forward1():
    for i in range(10):
        print(i, end=' ')
    print("done")

@autojit
def forward2():
    for i in range(1, 5):
        print(i, end=' ')
    print("done")

@autojit
def forward3():
    for i in range(5, 15, 3):
        print(i, end=' ')
    print("done")

@autojit
def backward1():
    for i in range(10, 2, -3):
        print(i, end=' ')
    print("done")

@autojit
def backward2():
    for i in range(1, 5, -1):
        print(i, end=' ')
    print("done")

@autojit
def backward3():
    for i in range(-5, -15, -3):
        print(i, end=' ')
    print("done")

@autojit
def empty_assign():
    i = 14
    for i in range(10, 4):
        pass
    print(i)

@autojit(warnstyle='simple')
def last_value():
    for i in range(10):
        pass

    print(i)

if __name__ == '__main__':
    # backward3()
    import numba
    numba.testing.testmod()



# -*- coding: utf-8 -*-

"""
Special compiler-recognized numba functions and attributes.
"""

from __future__ import print_function, division, absolute_import

__all__ = ['NULL', 'typeof', 'python', 'nopython', 'addressof']

import ctypes

from numba import error

#------------------------------------------------------------------------
# Pointers
#------------------------------------------------------------------------

class NumbaDotNULL(object):
    "NULL pointer"

NULL = NumbaDotNULL()


def addressof(obj, propagate=True):
    """
    Take the address of a compiled jit function.

    :param obj: the jit function
    :param write_unraisable: whether to write uncaught exceptions to stderr
    :param propagate: whether to always propagate exceptions

    :return: ctypes function pointer
    """
    from numba import numbawrapper

    if not propagate:
        raise ValueError("Writing unraisable exception is not yet supported")

    if not isinstance(obj, numbawrapper.numbafunction_type):
        raise TypeError("Object is not a jit function")

    if obj.lfunc_pointer is None:
        raise ValueError(
            "Jit function does not have pointer")

    ctypes_sig = obj.signature.to_ctypes()
    return ctypes.cast(obj.lfunc_pointer, ctypes_sig)

#------------------------------------------------------------------------
# Types
#------------------------------------------------------------------------

def typeof(variable):
    """
    Get the type of a variable.

    Used outside of Numba code, infers the type for the object.
    """
    from numba.environment import NumbaEnvironment
    context = NumbaEnvironment.get_environment().context
    return context.typemapper.from_python(variable)

#------------------------------------------------------------------------
# python/nopython context managers
#------------------------------------------------------------------------

class NoopContext(object):

    def __init__(self, name):
        self.name = name

    def __enter__(self, *args):
        return None

    def __exit__(self, *args):
        return None

    def __repr__(self):
        return self.name

python = NoopContext("python")
nopython = NoopContext("nopython")




import numpy as np
import numba
from numba import *
from numba.testing.test_support import autojit_py3doc

@jit
class Foo(object):
    def __init__(self, arg):
        self.arg = double(arg)

def test_typeof_pure(arg):
    """
    >>> test_typeof_pure(10)
    int
    >>> test_typeof_pure(10.0)
    double
    >>> print(test_typeof_pure(Foo(10)))
    <JitExtension Foo({'arg': double})>
    """
    return numba.typeof(arg)

@autojit_py3doc
def test_typeof_numba(a, b):
    """
    >>> test_typeof_numba(10, 11.0)
    21L
    >>> test_typeof_numba(11.0, 10)
    21.0
    >>> test_typeof_numba(np.arange(10), 1)
    array([ 1,  2,  3,  4,  5,  6,  7,  8,  9, 10])
    """
    return numba.typeof(a)(a + b)

@autojit
def test_typeof_numba2(arg):
    """
    >>> test_typeof_numba2(10)
    (10+0j)
    """
    x = 1 + 2j
    arg = numba.typeof(x)(arg)
    return numba.typeof(arg)(arg)

@autojit
def test_typeof_numba3(arg):
    """
    >>> print(test_typeof_numba3(10))
    int
    >>> print(test_typeof_numba3(Foo(10)))
    <JitExtension Foo({'arg': double})>
    """
    return numba.typeof(arg)

numba.testing.testmod()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import os
import sys
from itertools import ifilter
from functools import partial
import subprocess

from numba import PY3

import numba
root = os.path.dirname(os.path.abspath(numba.__file__))

# ______________________________________________________________________
# Test filtering

EXCLUDE_TEST_PACKAGES = [
    "numba.minivect",
    "numba.pyextensibletype",
    "numba.tests.broken_issues",
]

def make_path(root, predicate):
    "Call the predicate with a file path (e.g. numba/test/foo.py)"
    return lambda item: predicate(os.path.join(root, item))

def qualify_path(root, predicate):
    "Call the predicate with a dotted name (e.g. numba.tests.foo)"
    return make_path(root, lambda item: predicate(qualify_test_name(item)))


class Filter(object):
    def __init__(self, matcher=None):
        self.matcher = matcher

    def filter(self, root, dirs, files):
        matcher = make_path(root, self.matcher)
        return ifilter(matcher, dirs), ifilter(matcher, files)

class PackageFilter(Filter):
    def filter(self, root, dirs, files):
        matcher = qualify_path(root, self.matcher)
        return ifilter(matcher, dirs), files

class ModuleFilter(Filter):
    def filter(self, root, dirs, files):
        matcher = qualify_path(root, self.matcher)
        return dirs, ifilter(matcher, files)

class FileFilter(Filter):
    def filter(self, root, dirs, files):
        return dirs, [fn for fn in files if fn.endswith(".py")]

# ______________________________________________________________________
# Test discovery

class Walker(object):
    def __init__(self, root, filters):
        self.root = root
        self.filters = filters

    def walk(self):
        for root, dirs, files in os.walk(self.root):
            dirs[:], files[:] = apply_filters(root, dirs, files, self.filters)
            yield ([os.path.join(root, dir) for dir in dirs],
                   [os.path.join(root, fn) for fn in files])


def apply_filters(root, dirs, files, filters):
    for filter in filters:
        dirs, files = list(dirs), list(files)
        # print(filter, list(dirs), list(files))
        dirs, files = filter.filter(root, dirs, files)

    return dirs, files

def qualify_test_name(root):
    root, ext = os.path.splitext(root)
    qname = root.replace("/", ".").replace("\\", ".").replace(os.sep, ".") + "."
    offset = qname.rindex('numba.')
    return qname[offset:].rstrip(".")

def match(items, modname):
    return any(item in modname for item in items)

# ______________________________________________________________________
# Signal handling

def map_returncode_to_message(retcode):
    if retcode < 0:
        retcode = -retcode
        return signal_to_name.get(retcode, "Signal %d" % retcode)

    return ""

try:
    import signal
except ImportError:
    signal_to_name = {}
else:
    signal_to_name = dict((signal_code, signal_name)
                           for signal_name, signal_code in vars(signal).items()
                               if signal_name.startswith("SIG"))

# ______________________________________________________________________
# Test running

def test(whitelist=None, blacklist=None, print_failures_only=False):
    # FIXME
    # temporarily disable pycc test on win32
    if sys.platform.startswith('win32'):
        blacklist = ['test_pycc_tresult']

    # Make some test filters
    filters = [
        PackageFilter(lambda pkg: not any(
            pkg.startswith(p) for p in EXCLUDE_TEST_PACKAGES)),
        PackageFilter(lambda pkg: not pkg.endswith(".__pycache__")),
        ModuleFilter(lambda modname: modname.split('.')[-1].startswith("test_")),
        FileFilter(),
    ]

    if whitelist:
        filters.append(ModuleFilter(partial(match, whitelist)))

    if blacklist:
        filters.append(ModuleFilter(lambda item: not match(blacklist, item)))

    # Run tests
    runner = TestRunner(print_failures_only)
    run_tests(runner, filters)

    sys.stdout.write("ran test files: failed: (%d/%d)\n" % (runner.failed,
                                                            runner.ran))

    return 0 if runner.failed == 0 else 1

def run_tests(test_runner, filters, root=root):
    """
    Run tests:

        - Find tests in packages called 'tests'
        - Run any test files under a 'tests' package or a subpackage
    """
    testpkg_walker = Walker(root, filters)

    print("Running tests in %s" % os.path.join(root, "numba"))
    for testpkgs, _ in testpkg_walker.walk():
        for testpkg in testpkgs:
            if os.path.basename(testpkg) == "tests":
                # print("testdir:", testpkg)
                test_walker = Walker(testpkg, filters)
                for _, testfiles in test_walker.walk():
                    for testfile in testfiles:
                        # print("testfile:", testfile)
                        modname = qualify_test_name(testfile)
                        test_runner.run(modname)


class TestRunner(object):
    """
    Test runner used by runtests.py
    """

    def __init__(self, print_failures_only):
        self.ran = 0
        self.failed = 0
        self.print_failures_only = print_failures_only

    def run(self, modname):
        self.ran += 1
        if not self.print_failures_only:
            sys.stdout.write("%-70s" % (modname,))

        process = subprocess.Popen([sys.executable, '-m', modname],
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        out, err = process.communicate()

        if process.returncode == 0:
            if not self.print_failures_only:
                sys.stdout.write(" SUCCESS\n")
        else:
            if self.print_failures_only:
                sys.stdout.write("%-69s" % (modname,))

            sys.stdout.write(" FAILED:\n%79s\n" % map_returncode_to_message(
                                                        process.returncode))
            if PY3:
                out = str(out, encoding='UTF-8')
                err = str(err, encoding='UTF-8')
            sys.stdout.write(out)
            sys.stdout.write(err)
            sys.stdout.write("-" * 80)
            sys.stdout.write('\n')
            self.failed += 1

# ______________________________________________________________________
# Nose test running

def nose_run(module=None):
    import nose.config
    import __main__

    #os.environ["NOSE_EXCLUDE"] = "(test_all|test_all_noskip|.*compile_with_pycc.*|bytecode)"
    #os.environ["NOSE_VERBOSE"] = "4"

    result = nose.main()
    return len(result.errors), len(result.failures)



import os
from fnmatch import fnmatchcase
from distutils.util import convert_path
from distutils.core import setup
from Cython.Distutils import build_ext

from setupconfig import get_extensions

def find_packages(where='.', exclude=()):
    out = []
    stack=[(convert_path(where), '')]
    while stack:
        where, prefix = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where,name)
            if ('.' not in name and os.path.isdir(fn) and
                os.path.isfile(os.path.join(fn, '__init__.py'))
            ):
                out.append(prefix+name)
                stack.append((fn, prefix+name+'.'))
    for pat in list(exclude) + ['ez_setup', 'distribute_setup']:
        out = [item for item in out if not fnmatchcase(item, pat)]
    return out

root = os.path.dirname(os.path.abspath(__file__))
setup(cmdclass={'build_ext': build_ext},
      ext_modules=get_extensions(path_prefix=root),
      packages=find_packages())



"""
This modules creates extension types.

See also numba.extension_type_inference
"""

cimport cython
from numba._numba cimport *

import sys
import ctypes

import numba

ctypedef object (*tp_new_func)(PyObject *, PyObject *, PyObject *)
ctypedef void (*destructor)(PyObject *)
ctypedef int (*visitproc)(PyObject *, void *)

cdef extern from *:
    ctypedef struct PyTypeObject:
        tp_new_func tp_new
        destructor tp_dealloc
        int (*tp_traverse)(PyObject *o, visitproc v, void *a)
        int (*tp_clear)(PyObject *o)

        long tp_dictoffset
        Py_ssize_t tp_itemsize
        Py_ssize_t tp_basicsize
        PyTypeObject *tp_base

#------------------------------------------------------------------------
# Code to compute table offsets in the objects
#------------------------------------------------------------------------

cdef Py_uintptr_t align(Py_uintptr_t p, size_t alignment) nogil:
    "Align on a boundary"
    cdef size_t offset

    with cython.cdivision(True):
        offset = p % alignment

    if offset > 0:
        p += alignment - offset

    return p

cdef void *align_pointer(void *memory, size_t alignment) nogil:
    "Align pointer memory on a given boundary"
    return <void *> align(<Py_uintptr_t> memory, alignment)

def compute_vtab_offset(py_class):
    "Returns the vtab pointer offset in the object"
    offset = getattr(py_class, '__numba_vtab_offset', None)
    if offset:
        return offset

    cdef PyTypeObject *type_p = <PyTypeObject *> py_class
    return align(type_p.tp_basicsize, 8)

def compute_attrs_offset(py_class):
    "Returns the start of the attribute struct"
    offset = getattr(py_class, '__numba_attr_offset', None)
    if offset:
        return offset

    return align(compute_vtab_offset(py_class) + sizeof(void *), 8)

#------------------------------------------------------------------------
# Code to handle tp_dealloc, tp_traverse and tp_clear
#------------------------------------------------------------------------
# Currently we don't use this, since it seems to bother the ctypes struct
# in the object dict, but only if we access obj._numba_attrs._fields_
# We will want to generate descriptors to access attributes from Python
# directly instead of ctypes, in which case we need to set these functions
# on the extension type object.

def compute_object_offsets(ext_type, struct_offset):
    """
    Given an typesystem.ExtensionType, compute the offsets of object
    pointers, relative to the start of the attribute offsets.
    """
    offsets = []
    attr_struct = ext_type.attribute_table.to_struct()
    for field_name, field_type in attr_struct.fields:
        if field_type.is_object or field_type.is_array:
            offset = attr_struct.offsetof(field_name)
            offsets.append(struct_offset + offset)

    return offsets

cdef inline list getoffsets(type_obj):
    cdef list offsets = type_obj.__numba_object_offset
    return offsets

cdef inline void setnone(PyObject **slot):
    cdef PyObject *tmp

    tmp = slot[0]
    Py_INCREF(<PyObject *> None)
    slot[0] = <PyObject *> None
    Py_XDECREF(tmp)

cdef void tp_dealloc(PyObject *self):
    """
    Dealloc for numba extension types.

    This is set on the base numba extension type, since the offsets include
    the offsets of all superclasses. I.e. an instance of a subtype will use
    the deallocator of the supertype, which uses the offsets of the subtype.
    """
    type_obj = <object> self.ob_type
    cdef PyTypeObject *ob_type = <PyTypeObject *> type_obj
    cdef Py_ssize_t offset
    cdef PyObject **slot

    for offset in getoffsets(type_obj):
        slot = <PyObject **> ((<char *> <PyObject *> self) + offset)
        # Py_CLEAR(slot[0])
        setnone(slot) # Does ctypes collect and traverse for us?

    # Call base class dealloc
    if ob_type.tp_base.tp_dealloc != NULL:
        ob_type.tp_base.tp_dealloc(<PyObject *> self)

cdef int tp_clear(PyObject *self):
    """
    tp_clear for numba extension types.
    """
    cdef PyTypeObject *ob_type = <PyTypeObject *> self.ob_type

    # Let superclass clear its attributes
    if ob_type.tp_base.tp_clear != NULL:
        ob_type.tp_base.tp_clear(self)

    # Clear our own attributes
    cdef Py_ssize_t offset
    cdef PyObject **slot


    for offset in getoffsets(<object> ob_type):
        slot = <PyObject **> ((<char *> <PyObject *> self) + offset)
        setnone(slot)

    return 0

cdef int tp_traverse(PyObject *self, visitproc visit, void *arg):
    """
    tp_traverse for numba extension types.
    """
    cdef PyTypeObject *ob_type = <PyTypeObject *> self.ob_type
    cdef int result

    if ob_type.tp_base.tp_traverse != NULL:
        result = ob_type.tp_base.tp_traverse(self, visit, arg)
        if result:
            return result

    cdef Py_ssize_t offset
    cdef PyObject *obj

    for offset in getoffsets(<object> ob_type):
        obj = (<PyObject **> ((<char *> <PyObject *> self) + offset))[0]
        if obj != NULL:
            result = visit(obj, arg)
            if result:
                return result

    return 0

#------------------------------------------------------------------------
# Simple Vtable Wrappers
#------------------------------------------------------------------------

class StaticVtableWrapper(object):
    def __init__(self, vtable):
        self.vtable = vtable

    def get_vtable_pointer(self):
        vtab_p = ctypes.byref(self.vtable)
        return ctypes.cast(vtab_p, ctypes.c_void_p).value

cdef class DynamicVtableWrapper(object):

    cdef void *vtable_pointer
    cdef public object vtable

    def __init__(self, vtable):
        self.vtable = vtable
        self.write_vtable_pointer(vtable)

    def get_vtable_pointer(self):
        # Return a vtable **
        return <Py_uintptr_t> &self.vtable_pointer

    def write_vtable_pointer(self, new_table):
        self.vtable_pointer = <void *> <Py_uintptr_t> self.vtable.table_ptr

    def replace_vtable(self, new_vtable):
        old_vtable = self.vtable
        self.vtable = new_vtable
        self.write_vtable_pointer(new_vtable)

        # TODO: Manage keep alive of old_vtable until all consumers are done
        # TODO: with it

#------------------------------------------------------------------------
# Create Extension Type
#------------------------------------------------------------------------

def create_new_extension_type(metacls, name, bases, classdict,
                              exttype, vtable_wrapper):
    """
    Create an extension type from the given name, bases and classdict. Also
    takes a vtab struct minitype, and a attr_struct_type describing the
    object attributes.

    For static extension types:

            {
                PyObject_HEAD
                ...
                vtable *
                attribute1
                   ...
                attributeN
            }

        * The vtable is a contiguous block of memory holding function pointer

        * Attributes are put directly in the object

            * base classes must be compatible to allow a static ordering
              (we don't do C++-like dynamic offsets)

    For dynamic extension types:

            {
                PyObject_HEAD
                ...
                hash-based vtable *
                hash-based attribute table *
                attribute1
                   ...
                attributeN
            }

        * The vtable is a perfect hash-based vtable

        * Attributes are put directly in the object, BUT consumers may
          not rely on any ordering

              * Attribute locations are determined by a perfect hash-based
                table holding pointers to the attributes
    """
    cdef PyTypeObject *extclass_p
    cdef Py_ssize_t vtab_offset, attrs_offset

    orig_new = classdict.get('__new__', None)
    if orig_new:
        #
        orig_new = orig_new.__func__ # staticmethod is not callable!

    # __________________________________________________________________
    # Extension type constructor

    def new(cls, *args, **kwds):
        "Create a new object and patch it with a vtab"
        cdef PyObject *obj_p
        cdef void **vtab_location

        if orig_new is not None:
            new_func = orig_new
        else:
            assert issubclass(cls, extclass), (cls, extclass)
            new_func = super(extclass, cls).__new__

        if base_is_object:
            # Avoid warnings in py2.6 and errors in py3.x:
            #     DeprecationWarning: object.__new__() takes no parameters
            obj = new_func(cls)
        else:
            obj = new_func(cls, *args, **kwds)

        if (cls.__numba_vtab is not extclass.__numba_vtab or
                not isinstance(obj, cls)):
            # Subclass will set the vtab and attributes
            return obj

        # It is our responsibility to set the vtab and the ctypes attributes
        # Other fields are 0/NULL
        obj_p = <PyObject *> obj

        vtab_location = <void **> ((<char *> obj_p) + vtab_offset)
        vtab_location[0] = <void *> <Py_uintptr_t> cls.__numba_vtab_p

        attrs_pointer = (<Py_uintptr_t> obj_p) + attrs_offset
        obj._numba_attrs = ctypes.cast(attrs_pointer,
                                       cls.__numba_attributes_ctype)[0]

        return obj

    # __________________________________________________________________
    # Create extension type

    classdict['__new__'] = staticmethod(new)
    extclass = metacls(name, bases, classdict)
    assert isinstance(extclass, type)

    superclass_new = super(extclass, extclass).__new__
    cdef bint base_is_object = superclass_new.__self__ is object

    extclass_p = <PyTypeObject *> extclass

    # __________________________________________________________________
    # Update extension type

    attr_struct_type = exttype.attribute_table.to_struct()
    attr_struct_ctype = attr_struct_type.to_ctypes()

    vtab_offset = compute_vtab_offset(extclass)
    attrs_offset = compute_attrs_offset(extclass)

    upper = attrs_offset + ctypes.sizeof(attr_struct_ctype)
    upper = align(upper, 8)

    # print 'basicsize/vtab_offset/upper', extclass_p.tp_basicsize, lower, upper
    extclass_p.tp_basicsize = upper
    if extclass_p.tp_itemsize:
        raise NotImplementedError("Subclassing variable sized objects")

    # TODO: Figure out ctypes policy on tp_dealloc/tp_traverse/tp_clear
    # TODO: Don't use ctypes at all, but generate descriptors directly
    #if exttype.need_tp_dealloc:
    #    extclass_p.tp_dealloc = <destructor> tp_dealloc
    #    extclass_p.tp_clear = tp_clear
    #    extclass_p.tp_traverse = tp_traverse

    extclass.__numba_vtab_offset = vtab_offset
    extclass.__numba_obj_end = upper
    extclass.__numba_attr_offset = attrs_offset

    # Keep vtable alive
    if vtable_wrapper is None:
        # vtable_wrapper is None e.g. for the closure scope, which does
        # not need methods
        extclass.__numba_vtab = None
        extclass.__numba_vtab_p = 0
    else:
        extclass.__numba_vtab = vtable_wrapper
        extclass.__numba_vtab_p = vtable_wrapper.get_vtable_pointer()

    # __________________________________________________________________
    # Set ctypes attribute struct type, such that __new__ can create
    # _numba_attrs

    # TODO: Remove ctypes dependency and generate conversion descriptors

    extclass.__numba_attributes_ctype = attr_struct_type.pointer().to_ctypes()

    # __________________________________________________________________
    # Set exttype attribute

    extclass.__numba_ext_type = exttype
    extclass.exttype = exttype

    # __________________________________________________________________
    # Set offsets to objects (for tp_traverse, tp_clear, etc)

    offsets = getattr(extclass, '__numba_object_offset', [])
    offsets = offsets + compute_object_offsets(exttype, attrs_offset)
    extclass.__numba_object_offset = offsets

    return extclass




from numba import exportmany, export
from numba import decorators

def mult(a, b):
    return a * b

exportmany(['multf f4(f4,f4)', 'multi i4(i4,i4)'])(mult)
export('multc c16(c16,c16)')(mult)
export('mult f8(f8, f8)')(mult)



# -*- coding: utf-8 -*-

"""
This module defines ordering schemes for virtual methods and attributes.

If we use hash-based virtual (method/attribute) tables, we don't care about
the ordering. If we're using a C++ like virtual method/attribute table (like
normal Python extension types do for attributes), we need to have a layout
compatible with base classes (i.e. we may only add more attributes, but not
reorder any existing ones).
"""

from __future__ import print_function, division, absolute_import

from numba.traits import traits, Delegate
from numba import error

#------------------------------------------------------------------------
# Virtual Tables
#------------------------------------------------------------------------

@traits
class AbstractTable(object):

    # Ordered attribute names
    attributes = None

    # Dict mapping attribute names to attribute entities
    attrdict = None

    py_class = Delegate('table')

    def __init__(self, table):
        self.table = table

    @property
    def parents(self):
        cls = type(self)
        return list(map(cls, self.table.parents))

@traits
class VTable(AbstractTable):

    attributes = Delegate('table', 'methodnames')
    attrdict = Delegate('table', 'methoddict')

@traits
class AttributeTable(AbstractTable):

    attributes = Delegate('table', 'attributes')
    attrdict = Delegate('table', 'attributedict')

#------------------------------------------------------------------------
# Table Entry Ordering (Virtual Method / Attribute Ordering)
#------------------------------------------------------------------------

def sort_parents(table):
    "Sort parent tables by size"
    return sorted(table.parents, key=lambda tab: len(tab.attrdict))

def unordered(table):
    "Return table entities in a random order"
    return list(table.attrdict)

def extending(table):
    """
    Order the table entities according to the given parent tables, i.e.
    we can only extend existing tables.
    """
    if not table.parents:
        return unordered(table)

    parents = sort_parents(table)
    biggest_table = parents[-1]

    appending_attributes = set(table.attrdict) - set(biggest_table.attributes)
    return biggest_table.attributes + list(appending_attributes)


# ______________________________________________________________________
# Validate Table Ordering

def validate_extending_order_compatibility(table):
    parents = sort_parents(table)
    tables = parents + [table]

    for table_smaller, table_bigger in zip(tables, tables[1:]):
        names1 = table_smaller.attributes
        names2 = table_bigger.attributes[:len(table_smaller.attributes)]

        if names1 != names2:
            raise error.NumbaError(
                "Cannot create compatible attribute or method ordering for "
                "base classes '%s' and '%s'" % (
                    table_smaller.py_class.__name__,
                    table_bigger.py_class.__name__))




import numba; from numba import *; from numba.error import NumbaError
from numba.testing.test_support import rewrite_doc
@autojit
def error1():
    def inner():
        pass

@autojit
def error2():
    @autojit
    def inner():
        pass

@autojit
def error3():
    inner(10, 20, 30)

    @jit(restype=void, argtypes=[int_, int_, int_])
    def inner(a, b, c):
        print(str(a) + ' ' + str(b) + ' ' + str(c))

@autojit
def error4():
    @jit(restype=void, argtypes=[int_, int_, int_])
    def inner(a, b, c):
        print(str(a) + ' ' + str(b) + ' ' + str(c))

    inner(10, a=20, b=30, c=40)

@autojit
def error5():
    @jit(restype=void, argtypes=[int_, int_, int_])
    def inner(a, b, c):
        print(str(a) + ' ' + str(b) + ' ' + str(c))

    inner(10, a=20, b=30)

@autojit(warnstyle='simple')
def closure1():
    a = 10
    @jit(restype=void, argtypes=[int_])
    def inner(arg):
        print(str(arg))

    return inner

@autojit
def closure2():
    a = 10
    @jit(restype=void, argtypes=[int_])
    def inner(arg):
        print(str(arg))

    inner(arg=a)

@autojit
def closure3():
    a = 10

    @jit('void()')
    def inner():
        print(a)

    a = 12
    inner()

@autojit
def closure4():
    a = 10

    @jit('void()')
    def inner():
        print(a)

    a = 12
    return inner

@autojit
def nested_closure():
    a = 20
    b = 21

    @jit('object_()')
    def c1():
        b = 10
        @jit('void()')
        def c2():
            print(str(a) + ' ' + str(b))

        return c2

    @jit('void()')
    def c3():
        print(b)

    return c1, c3

__doc__ = """
>>> error1()
Traceback (most recent call last):
    ...
NumbaError: 5:4: Closure must be decorated with 'jit' or 'autojit'
>>> error2()
Traceback (most recent call last):
    ...
NumbaError: 10:5: Dynamic closures not yet supported, use @jit
>>> error3()
Traceback (most recent call last):
    ...
NumbaError: 16:4: local variable 'inner' referenced before assignment
>>> error4()
Traceback (most recent call last):
    ...
NumbaError: 28:4: Expected 3 arguments, got 4
>>> error5()
Traceback (most recent call last):
    ...
NumbaError: Got multiple values for positional argument 'a'

Test closures

>>> closure1().__name__
Warning 40:4: Unused variable 'a'
'inner'
>>> closure1()()
Traceback (most recent call last):
    ...
TypeError: function takes exactly 1 argument (0 given)
>>> closure1()(object())
Traceback (most recent call last):
    ...
TypeError: an integer is required
>>> closure1()(10.0)
10
>>> closure2()
10
>>> closure3()
12
>>> func = closure4()
>>> print(func.__name__)
inner
>>> field, = func.__closure__._numba_attrs._fields_
>>> import ctypes
>>> print((field[0], field[1] == ctypes.c_int))
('a', True)
>>> print(func.__closure__._numba_attrs.a)
12
>>> func()
12

>>> c1, c3 = nested_closure()
>>> c1.__name__
'c1'
>>> c3.__name__
'c3'
>>> c1().__name__
'c2'
>>> c1()()
20 10
>>> c3()
21
"""





@autojit
def closure_arg(a):
    @jit('object_(object_)')
    def closure1(b):
        print(str(a) + ' ' + str(b))
        x = 10 + int_(b)
        @jit('object_(object_)')
        def closure2(c):
            print(str(a) + ' ' + str(b) + ' ' + str(c) + ' ' + str(x))
            y = double(x) + double(c)
            @jit('void(object_)')
            def closure3(d):
                print(str(a) + ' ' + str(b) + ' ' + str(c) + ' ' + str(d) + ' ' + str(x) + ' ' + str(y))
            return closure3
        return closure2
    return closure1

__doc__ += \
"""
>>> closure1 = closure_arg(1)
>>> closure1.__name__
'closure1'

>>> closure2_1 = closure1(2)
1 2
>>> closure2_1.__name__
'closure2'
>>> closure2_2 = closure1(3)
1 3
>>> closure2_2.__name__
'closure2'

>>> closure3_1 = closure2_1(4)
1 2 4 12
>>> closure3_1.__name__
'closure3'
>>> closure3_2 = closure2_2(5)
1 3 5 13
>>> closure3_2.__name__
'closure3'

>>> closure3_1(6)
1 2 4 6 12 16.0
>>> closure3_2(7)
1 3 5 7 13 18.0
"""


@autojit
def closure_arg_simple(a):
    @jit('object_(object_)')
    def inner(b):
        print(str(a) + ' ' + str(b))
        @jit('void(object_)')
        def inner_inner(c):
            print(str(a) + ' ' + str(b) + ' ' + str(c))
        return inner_inner
    return inner

__doc__ += """
>>> closure_arg_simple(10)(20)(30)
10 20
10 20 30
"""

@autojit
def closure_skip_level(a):
    @jit('object_()')
    def inner():
        @jit('void()')
        def inner_inner():
            print(str(a))
        return inner_inner
    return inner

__doc__ += """
>>> closure_skip_level(10)()()
10
"""

@autojit
def objects(s):
    @jit('object_()')
    def inner():
        return s.upper()
    return inner

__doc__ += """
>>> objects("hello")()
'HELLO'
"""

@autojit
def wrong_signature(s):
    @jit('object_(object_)')
    def inner():
        return s.upper()
    return inner

__doc__ += """
>>> try_(wrong_signature, "foo")
--------------------- Numba Encountered Errors or Warnings ---------------------
    @jit('object_(object_)')
-----^
Error 262:5: Expected 1 arguments type(s), got 0
--------------------------------------------------------------------------------
NumbaError: 262:5: Expected 1 arguments type(s), got 0
"""



@autojit
def wrong_restype():
    @jit('object_()')
    def inner():
        pass
    return inner

__doc__ += """
>>> try_(wrong_restype)
--------------------- Numba Encountered Errors or Warnings ---------------------
    @jit('object_()')
----^
Error 281:4: Function with non-void return does not return a value
--------------------------------------------------------------------------------
NumbaError: 281:4: Function with non-void return does not return a value
"""

#
### Test signatures like @double(object_)
#
@autojit
def signature_dec():
    @object_()
    def inner():
        return "hello"
    return inner

__doc__ += """
>>> signature_dec()()
'hello'
"""

@autojit
def wrong_signature2(s):
    @object_(object_)
    def inner():
        return s.upper()
    return inner

__doc__ += """
>>> try_(wrong_signature2, "foo")
--------------------- Numba Encountered Errors or Warnings ---------------------
    @object_(object_)
-----^
Error 313:5: Expected 1 arguments type(s), got 0
--------------------------------------------------------------------------------
NumbaError: 313:5: Expected 1 arguments type(s), got 0
"""

@autojit
def get_closure(arg):
    @void()
    def closure():
        print(arg)

    closure()
    return closure

@autojit
def test_call_closure():
    closure = get_closure(10.0)
    closure() # TODO: This still goes through the object layer, amend

__doc__ += """
>>> test_call_closure()
10.0
10.0
"""

@autojit
def test_call_closure_from_closure():
    closure = get_closure(10.0)
    @void()
    def inner():
        closure()
    return inner

__doc__ += """
>>> test_call_closure_from_closure()()
10.0
10.0
"""

@autojit
def test_closure_loop():
    """
    >>> test_closure_loop()
    0 3
    1 3
    2 3
    <BLANKLINE>
    0 3
    1 3
    2 3
    """
    cellvar = 3

    @jit(void())
    def inner():
        for i in range(cellvar):
            print(str(i) + ' ' + str(cellvar))

        print('')

        for i in range(cellvar):
            for j in range(cellvar):
                if i == j:
                    print(str(i) + ' ' + str(cellvar))

    inner()

@autojit(locals=dict(var=int_), warn=False)
def test_closure_outer_locals():
    """
    >>> test_closure_outer_locals()
    """
    var = 10
    @jit(void())
    def inner():
        var = "hello"

    inner()

#__doc__ = rewrite_doc(__doc__)

def try_(func, *args):
    try:
        func(*args)
    except NumbaError as e:
        print("%s%s: %s" % ('numba.error.' if numba.PY3 else '',
                            type(e).__name__, e))

if __name__ == '__main__':
#    closure1 = closure_arg(1)
    # print closure1.__name__
    # closure1(10)
    # test_call_closure()
    # closure4()
    # signature_dec()()
    # test_closure_outer_locals()
    # test_closure_loop()
    # test_closure_outer_locals()
#    test_call_closure_from_closure()()
    # wrong_restype()
    import numba
    numba.testing.testmod()



# -*- coding: utf-8 -*-

"""
Test hash-based virtual method tables.
"""

from __future__ import print_function, division, absolute_import

import itertools

import numba as nb
from numba import *
from numba.minivect.minitypes import FunctionType
from numba.exttypes import virtual
from numba.exttypes import methodtable
from numba.exttypes.signatures import Method
from numba.testing.test_support import parametrize, main

#------------------------------------------------------------------------
# Signature enumeration
#------------------------------------------------------------------------

class py_class(object):
    pass

def myfunc1(a):
    pass

def myfunc2(a, b):
    pass

def myfunc3(a, b, c):
    pass

types = list(nb.numeric) + [object_]

array_types = [t[:] for t in types]
array_types += [t[:, :] for t in types]
array_types += [t[:, :, :] for t in types]

all_types = types + array_types

def method(func, name, sig):
    return Method(func, name, sig, False, False)

make_methods1 = lambda: [
    method(myfunc1, 'method', FunctionType(argtype, [argtype]))
        for argtype in all_types]

make_methods2 = lambda: [
    method(myfunc2, 'method', FunctionType(argtype1, [argtype1, argtype2]))
        for argtype1, argtype2 in itertools.product(all_types, all_types)]

#------------------------------------------------------------------------
# Table building
#------------------------------------------------------------------------

def make_table(methods):
    table = methodtable.VTabType(py_class, [])
    table.create_method_ordering()

    for i, method in enumerate(methods):
        key = method.name, method.signature.args
        method.lfunc_pointer = i
        table.specialized_methods[key] = method

    assert len(methods) == len(table.specialized_methods)

    return table

def make_hashtable(methods):
    table = make_table(methods)
    hashtable = virtual.build_hashing_vtab(table)
    return hashtable

#------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------

@parametrize(make_methods1(), make_methods2()[:6000])
def test_specializations(methods):
    hashtable = make_hashtable(methods)
    # print(hashtable)

    for i, method in enumerate(methods):
        key = virtual.sep201_signature_string(method.signature, method.name)
        assert hashtable.find_method(key), (i, method, key)


if __name__ == '__main__':
    # test_specializations(make_methods2())
    main()



# -*- coding: utf-8 -*-
"""An implementation of the Zephyr Abstract Syntax Definition Language.

See http://asdl.sourceforge.net/ and
http://www.cs.princeton.edu/research/techreps/TR-554-97

Only supports top level module decl, not view.  I'm guessing that view
is intended to support the browser and I'm not interested in the
browser.

Changes for Python: Add support for module versions
"""
from __future__ import print_function, division, absolute_import

import os
import traceback

from . import spark

class Token(object):
    # spark seems to dispatch in the parser based on a token's
    # type attribute
    def __init__(self, type, lineno):
        self.type = type
        self.lineno = lineno

    def __str__(self):
        return self.type

    def __repr__(self):
        return str(self)

class Id(Token):
    def __init__(self, value, lineno):
        self.type = 'Id'
        self.value = value
        self.lineno = lineno

    def __str__(self):
        return self.value

class String(Token):
    def __init__(self, value, lineno):
        self.type = 'String'
        self.value = value
        self.lineno = lineno

class ASDLSyntaxError(Exception):

    def __init__(self, lineno, token=None, msg=None):
        self.lineno = lineno
        self.token = token
        self.msg = msg

    def __str__(self):
        if self.msg is None:
            return "Error at '%s', line %d" % (self.token, self.lineno)
        else:
            return "%s, line %d" % (self.msg, self.lineno)

class ASDLScanner(spark.GenericScanner, object):

    def tokenize(self, input):
        self.rv = []
        self.lineno = 1
        super(ASDLScanner, self).tokenize(input)
        return self.rv

    def t_id(self, s):
        r"[\w\.]+"
        # XXX doesn't distinguish upper vs. lower, which is
        # significant for ASDL.
        self.rv.append(Id(s, self.lineno))

    def t_string(self, s):
        r'"[^"]*"'
        self.rv.append(String(s, self.lineno))

    def t_xxx(self, s): # not sure what this production means
        r"<="
        self.rv.append(Token(s, self.lineno))

    def t_punctuation(self, s):
        r"[\{\}\*\=\|\(\)\,\?\:]"
        self.rv.append(Token(s, self.lineno))

    def t_comment(self, s):
        r"\-\-[^\n]*"
        pass

    def t_newline(self, s):
        r"\n"
        self.lineno += 1

    def t_whitespace(self, s):
        r"[ \t]+"
        pass

    def t_default(self, s):
        r" . +"
        raise ValueError("unmatched input: %r" % s)

class ASDLParser(spark.GenericParser, object):
    def __init__(self):
        super(ASDLParser, self).__init__("module")

    def typestring(self, tok):
        return tok.type

    def error(self, tok):
        raise ASDLSyntaxError(tok.lineno, tok)

    def p_module_0(self, xxx_todo_changeme):
        " module ::= Id Id version { } "
        (module, name, version, _0, _1) = xxx_todo_changeme
        if module.value != "module":
            raise ASDLSyntaxError(module.lineno,
                                  msg="expected 'module', found %s" % module)
        return Module(name, None, version)

    def p_module(self, xxx_todo_changeme1):
        " module ::= Id Id version { definitions } "
        (module, name, version, _0, definitions, _1) = xxx_todo_changeme1
        if module.value != "module":
            raise ASDLSyntaxError(module.lineno,
                                  msg="expected 'module', found %s" % module)
        return Module(name, definitions, version)

    def p_version(self, xxx_todo_changeme2):
        "version ::= Id String"
        (version, V) = xxx_todo_changeme2
        if version.value != "version":
            raise ASDLSyntaxError(version.lineno,
                                msg="expected 'version', found %" % version)
        return V

    def p_definition_0(self, xxx_todo_changeme3):
        " definitions ::= definition "
        (definition,) = xxx_todo_changeme3
        return definition

    def p_definition_1(self, xxx_todo_changeme4):
        " definitions ::= definition definitions "
        (definitions, definition) = xxx_todo_changeme4
        return definitions + definition

    def p_definition(self, xxx_todo_changeme5):
        " definition ::= Id = type "
        (id, _, type) = xxx_todo_changeme5
        return [Type(id, type)]

    def p_type_0(self, xxx_todo_changeme6):
        " type ::= product "
        (product,) = xxx_todo_changeme6
        return product

    def p_type_1(self, xxx_todo_changeme7):
        " type ::= sum "
        (sum,) = xxx_todo_changeme7
        return Sum(sum)

    def p_type_2(self, xxx_todo_changeme8):
        " type ::= sum Id ( fields ) "
        (sum, id, _0, attributes, _1) = xxx_todo_changeme8
        if id.value != "attributes":
            raise ASDLSyntaxError(id.lineno,
                                  msg="expected attributes, found %s" % id)
        if attributes:
            attributes.reverse()
        return Sum(sum, attributes)

    def p_product(self, xxx_todo_changeme9):
        " product ::= ( fields ) "
        (_0, fields, _1) = xxx_todo_changeme9
        fields.reverse()
        return Product(fields)

    def p_sum_0(self, xxx_todo_changeme10):
        " sum ::= constructor "
        (constructor,) = xxx_todo_changeme10
        return [constructor]

    def p_sum_1(self, xxx_todo_changeme11):
        " sum ::= constructor | sum "
        (constructor, _, sum) = xxx_todo_changeme11
        return [constructor] + sum

    def p_sum_2(self, xxx_todo_changeme12):
        " sum ::= constructor | sum "
        (constructor, _, sum) = xxx_todo_changeme12
        return [constructor] + sum

    def p_constructor_0(self, xxx_todo_changeme13):
        " constructor ::= Id "
        (id,) = xxx_todo_changeme13
        return Constructor(id)

    def p_constructor_1(self, xxx_todo_changeme14):
        " constructor ::= Id ( fields ) "
        (id, _0, fields, _1) = xxx_todo_changeme14
        fields.reverse()
        return Constructor(id, fields)

    def p_fields_0(self, xxx_todo_changeme15):
        " fields ::= field "
        (field,) = xxx_todo_changeme15
        return [field]

    def p_fields_1(self, xxx_todo_changeme16):
        " fields ::= field , fields "
        (field, _, fields) = xxx_todo_changeme16
        return fields + [field]

    def p_field_0(self, xxx_todo_changeme17):
        " field ::= Id "
        (type,) = xxx_todo_changeme17
        return Field(type)

    def p_field_1(self, xxx_todo_changeme18):
        " field ::= Id Id "
        (type, name) = xxx_todo_changeme18
        return Field(type, name)

    def p_field_2(self, xxx_todo_changeme19):
        " field ::= Id * Id "
        (type, _, name) = xxx_todo_changeme19
        return Field(type, name, seq=True)

    def p_field_3(self, xxx_todo_changeme20):
        " field ::= Id ? Id "
        (type, _, name) = xxx_todo_changeme20
        return Field(type, name, opt=True)

    def p_field_4(self, xxx_todo_changeme21):
        " field ::= Id * "
        (type, _) = xxx_todo_changeme21
        return Field(type, seq=True)

    def p_field_5(self, xxx_todo_changeme22):
        " field ::= Id ? "
        (type, _) = xxx_todo_changeme22
        return Field(type, opt=True)

builtin_types = ("identifier", "string", "int", "bool", "object")

# below is a collection of classes to capture the AST of an AST :-)
# not sure if any of the methods are useful yet, but I'm adding them
# piecemeal as they seem helpful

class AST(object):
    pass # a marker class

class Module(AST):
    def __init__(self, name, dfns, version):
        self.name = name
        self.dfns = dfns
        self.version = version
        self.types = {} # maps type name to value (from dfns)
        for type in dfns:
            self.types[type.name.value] = type.value

    def __repr__(self):
        return "Module(%s, %s)" % (self.name, self.dfns)

class Type(AST):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __repr__(self):
        return "Type(%s, %s)" % (self.name, self.value)

class Constructor(AST):
    def __init__(self, name, fields=None):
        self.name = name
        self.fields = fields or []

    def __repr__(self):
        return "Constructor(%s, %s)" % (self.name, self.fields)

class Field(AST):
    def __init__(self, type, name=None, seq=False, opt=False):
        self.type = type
        self.name = name
        self.seq = seq
        self.opt = opt

    def __repr__(self):
        if self.seq:
            extra = ", seq=True"
        elif self.opt:
            extra = ", opt=True"
        else:
            extra = ""
        if self.name is None:
            return "Field(%s%s)" % (self.type, extra)
        else:
            return "Field(%s, %s%s)" % (self.type, self.name, extra)

class Sum(AST):
    def __init__(self, types, attributes=None):
        self.types = types
        self.attributes = attributes or []

    def __repr__(self):
        if self.attributes is None:
            return "Sum(%s)" % self.types
        else:
            return "Sum(%s, %s)" % (self.types, self.attributes)

class Product(AST):
    def __init__(self, fields):
        self.fields = fields

    def __repr__(self):
        return "Product(%s)" % self.fields

class VisitorBase(object):

    def __init__(self, skip=False):
        self.cache = {}
        self.skip = skip

    def visit(self, object, *args):
        meth = self._dispatch(object)
        if meth is None:
            return
        try:
            meth(object, *args)
        except Exception as err:
            print(("Error visiting", repr(object)))
            print(err)
            traceback.print_exc()
            # XXX hack
            if hasattr(self, 'file'):
                self.file.flush()
            os._exit(1)

    def _dispatch(self, object):
        assert isinstance(object, AST), repr(object)
        klass = object.__class__
        meth = self.cache.get(klass)
        if meth is None:
            methname = "visit" + klass.__name__
            if self.skip:
                meth = getattr(self, methname, None)
            else:
                meth = getattr(self, methname)
            self.cache[klass] = meth
        return meth

class Check(VisitorBase):

    def __init__(self):
        super(Check, self).__init__(skip=True)
        self.cons = {}
        self.errors = 0
        self.types = {}

    def visitModule(self, mod):
        for dfn in mod.dfns:
            self.visit(dfn)

    def visitType(self, type):
        self.visit(type.value, str(type.name))

    def visitSum(self, sum, name):
        for t in sum.types:
            self.visit(t, name)

    def visitConstructor(self, cons, name):
        key = str(cons.name)
        conflict = self.cons.get(key)
        if conflict is None:
            self.cons[key] = name
        else:
            print(("Redefinition of constructor %s" % key))
            print(("Defined in %s and %s" % (conflict, name)))
            self.errors += 1
        for f in cons.fields:
            self.visit(f, key)

    def visitField(self, field, name):
        key = str(field.type)
        l = self.types.setdefault(key, [])
        l.append(name)

    def visitProduct(self, prod, name):
        for f in prod.fields:
            self.visit(f, name)

def check(mod):
    v = Check()
    v.visit(mod)

    for t in v.types:
        if t not in mod.types and not t in builtin_types:
            v.errors += 1
            uses = ", ".join(v.types[t])
            print(("Undefined type %s, used in %s" % (t, uses)))

    return not v.errors

def parse(file):
    scanner = ASDLScanner()
    parser = ASDLParser()

    buf = open(file).read()
    tokens = scanner.tokenize(buf)
    try:
        return parser.parse(tokens)
    except ASDLSyntaxError as err:
        print(err)
        lines = buf.split("\n")
        print((lines[err.lineno - 1])) # lines starts at 0, files at 1

if __name__ == "__main__":
    import glob
    import sys

    if len(sys.argv) > 1:
        files = sys.argv[1:]
    else:
        testdir = "tests"
        files = glob.glob(testdir + "/*.asdl")

    for file in files:
        print(file)
        mod = parse(file)
        print(("module", mod.name))
        print((len(mod.dfns), "definitions"))
        if not check(mod):
            print("Check failed")
        else:
            for dfn in mod.dfns:
                print((dfn.type))



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from numba.nodes import *
import numba.nodes

class CheckErrorNode(ExprNode):
    """
    Check for an exception.

        badval: if this value is returned, propagate an error
        goodval: if this value is not returned, propagate an error

    If exc_type, exc_msg and optionally exc_args are given, an error is
    raised instead of propagating it.

    See RaiseNode for the exc_* arguments.
    """

    _fields = ['return_value', 'badval', 'raise_node']

    def __init__(self, return_value, badval=None, goodval=None,
                 exc_type=None, exc_msg=None, exc_args=None,
                 **kwargs):
        super(CheckErrorNode, self).__init__(**kwargs)
        self.return_value = return_value

        if badval is not None and not isinstance(badval, ast.AST):
            badval = ConstNode(badval, return_value.type)
        if goodval is not None and not isinstance(goodval, ast.AST):
            goodval = ConstNode(goodval, return_value.type)

        self.badval = badval
        self.goodval = goodval

        self.raise_node = RaiseNode(exc_type, exc_msg, exc_args)

class RaiseNode(ExprNode):
    """
    Raise an exception.

        exception_type: The Python exception type

        exc_type: The Python exception as an AST node
            May be passed in as a Python exception type

        exc_msg: The message to print as an AST node
            May be passed in as a string

        exc_args: If given, must be an list of AST nodes representing the
                  arguments to PyErr_Format (matching the format specifiers
                  at runtime in exc_msg)
    """

    _fields = ['exc_type', 'exc_msg', 'exc_args']

    def __init__(self, exc_type, exc_msg, exc_args=None, print_on_trap=True,
                 **kwargs):
        super(RaiseNode, self).__init__(**kwargs)
        self.exception_type = None
        if isinstance(exc_type, type) and issubclass(exc_type, BaseException):
            self.exception_type = exc_type
            exc_type = const(exc_type, object_)
        if isinstance(exc_msg, (str, unicode)):
            exc_msg = const(exc_msg, c_string_type)

        self.exc_type = exc_type
        self.exc_msg = exc_msg
        self.exc_args = exc_args

        self.print_on_trap = print_on_trap

class PropagateNode(ExprNode):
    """
    Propagate an exception (jump to the error label). This is resolved
    at code generation time and can be generated at any moment.
    """

class PyErr_OccurredNode(ExprNode):
    """
    Check for a set Python exception using PyErr_Occurred().

    Can be set any time after type inference. This node is resolved during
    late specialization.
    """

    # TODO: support checking for (value == badval && PyErr_Occurred()) for
    #       efficiency

    _fields = ['node']

    def __init__(self, node):
        self.node = node
        self.variable = node.variable
        self.type = node.type



# -*- coding: UTF-8 -*-
"""
numba --annotate

Adapted from Cython/Compiler/Annotate.py
"""
from __future__ import print_function, division, absolute_import

# Note: Work in progress

import os
import re
import time
import codecs
from StringIO import StringIO

import numba
from numba import utils
from numba import visitors

class AnnotationVisitor(visitors.NumbaVisitor):
    """
    Annotate Python source to produce a static webpage showing where
    code is using the Python C API.

    Runs somewhere after type inference.
    """

    def __init__(self, *args, **kwargs):
        super(AnnotationVisitor, self).__init__(*args, **kwargs)
        self.annotations = []

    def visit_Name(self, node):
        return AnnotationItem(node, "py_code", size=len(node.id))



# need one-characters subsitutions (for now) so offsets aren't off
special_chars = [(u'<', u'\xF0', u'&lt;'),
                 (u'>', u'\xF1', u'&gt;'),
                 (u'&', u'\xF2', u'&amp;')]

class AnnotationCodeWriter(object):

    def __init__(self):
        self.annotation_buffer = StringIO()
        self.annotations = []
        self.last_pos = None
        self.code = {}

    def write(self, s):
        self.annotation_buffer.write(s)

    def mark_pos(self, pos):
        if self.last_pos:
            pos_code = self.code.setdefault(self.last_pos[0].filename,{})
            code = pos_code.get(self.last_pos[1], "")
            pos_code[self.last_pos[1]] = code + self.annotation_buffer.getvalue()
        self.annotation_buffer = StringIO()
        self.last_pos = pos

    def annotate(self, pos, item):
        self.annotations.append((pos, item))

    def save_annotation(self, source_filename, target_filename):
        self.mark_pos(None)
        f = open(source_filename)
        lines = f.readlines()
        for k in range(len(lines)):
            line = lines[k]
            for c, cc, html in special_chars:
                line = line.replace(c, cc)
            lines[k] = line
        f.close()
        all = []
        if False:
            for pos, item in self.annotations:
                if pos[0].filename == source_filename:
                    start = item.start()
                    size, end = item.end()
                    if size:
                        all.append((pos, start))
                        all.append(((source_filename, pos[1], pos[2]+size), end))
                    else:
                        all.append((pos, start+end))

        all.sort()
        all.reverse()
        for pos, item in all:
            _, line_no, col = pos
            line_no -= 1
            col += 1
            line = lines[line_no]
            lines[line_no] = line[:col] + item + line[col:]

        html_filename = os.path.splitext(target_filename)[0] + ".html"
        f = codecs.open(html_filename, "w", encoding="UTF-8")
        watermark = u'<!-- Generated by Numba %s on %s -->\n' % (
                                numba.__version__, time.asctime())
        f.write(watermark)
        f.write(u'<html>\n')
        f.write(u"""
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
<style type="text/css">

body { font-family: courier; font-size: 12; }

.tag  {  }
.line { margin: 0em }

</style>
<script>
function toggleDiv(id) {
    theDiv = document.getElementById(id);
    if (theDiv.style.display == 'none') theDiv.style.display = 'block';
    else theDiv.style.display = 'none';
}
</script>
</head>
        """)
        f.write(u'<body>\n')
        f.write(watermark)
        c_file = utils.decode_filename(os.path.basename(target_filename))
        f.write(u'<p>Raw output: <a href="%s">%s</a>\n' % (c_file, c_file))
        k = 0

        py_c_api = re.compile(u'(Py[A-Z][a-z]+_[A-Z][a-z][A-Za-z_]+)\(')
        py_marco_api = re.compile(u'(Py[A-Z][a-z]+_[A-Z][A-Z_]+)\(')
        pyx_c_api = re.compile(u'(__Pyx_[A-Z][a-z_][A-Za-z_]+)\(')
        pyx_macro_api = re.compile(u'(__Pyx_[A-Z][A-Z_]+)\(')
        error_goto = re.compile(ur'((; *if .*)? \{__pyx_filename = .*goto __pyx_L\w+;\})')
        refnanny = re.compile(u'(__Pyx_X?(GOT|GIVE)REF|__Pyx_RefNanny[A-Za-z]+)')

        code_source_file = self.code[source_filename]
        for line in lines:

            k += 1
            try:
                code = code_source_file[k]
            except KeyError:
                code = ''

            code = code.replace('<', '<code><</code>')

            code, py_c_api_calls = py_c_api.subn(ur"<span class='py_c_api'>\1</span>(", code)
            code, pyx_c_api_calls = pyx_c_api.subn(ur"<span class='pyx_c_api'>\1</span>(", code)
            code, py_macro_api_calls = py_marco_api.subn(ur"<span class='py_macro_api'>\1</span>(", code)
            code, pyx_macro_api_calls = pyx_macro_api.subn(ur"<span class='pyx_macro_api'>\1</span>(", code)
            code, refnanny_calls = refnanny.subn(ur"<span class='refnanny'>\1</span>", code)
            code, error_goto_calls = error_goto.subn(ur"<span class='error_goto'>\1</span>", code)

            code = code.replace(u"<span class='error_goto'>;", u";<span class='error_goto'>")

            score = 5*py_c_api_calls + 2*pyx_c_api_calls + py_macro_api_calls + pyx_macro_api_calls - refnanny_calls
            color = u"FFFF%02x" % int(255/(1+score/10.0))
            f.write(u"<pre class='line' style='background-color: #%s' onclick='toggleDiv(\"line%s\")'>" % (color, k))

            f.write(u" %d: " % k)
            for c, cc, html in special_chars:
                line = line.replace(cc, html)
            f.write(line.rstrip())

            f.write(u'</pre>\n')
            code = re.sub(line_pos_comment, '', code) # inline annotations are redundant
            f.write(u"<pre id='line%s' class='code' style='background-color: #%s'>%s</pre>" % (k, color, code))
        f.write(u'</body></html>\n')
        f.close()


# TODO: make this cleaner
def escape(raw_string):
    raw_string = raw_string.replace(u"\'", ur"&#146;")
    raw_string = raw_string.replace(u'\"', ur'&quot;')
    raw_string = raw_string.replace(u'\n', ur'<br>\n')
    raw_string = raw_string.replace(u'\t', ur'\t')
    return raw_string

styles = {
    "py_code": "{}",
    "code": ("{ font-size: 9; color: #444444; display: none; " # LLVM code
             "margin-left: 20px; }"),
    "py_c_api": "{ color: red; }",
    "error_goto": "{ color: #FFA000; }",
    "coerce": ("{ color: #008000; " # (object <-> native coercion)
               "border: 1px dotted #008000 }"),
    "py_attr": "{ color: #FF0000; font-weight: bold; }",
    "c_attr": "{ color: #0000FF; }",
    "py_call": "{ color: #FF0000; font-weight: bold; }",
    "c_call": "{ color: #0000FF; }",
    "line": "{ margin: 0em }",
}

titles = {
    "code":         "LLVM Code",
    "py_c_api":     "Python C API",
    "error_goto":   "Error Checking",
    "coerce":       "Coercion to/from object",
    "py_attr":      "Python Attribute Access",
    "c_attr":       "Native Attribute Access",
    "py_call":      "Python Function Call",
    "c_call":       "Native Call",
}

class AnnotationItem(object):
    """
    Annotation of some code.

    Style is one of the following:

        * py_code           (Python Code)
        * code              (LLVM code)
        * py_c_api
        * py_macro_api      (NA)
        * error_goto        (error checking)
        * coerce            (object <-> native coercion)
        * py_attr
        * c_attr
        * py_call
        * c_call
    """

    def __init__(self, node, style, tag="", size=0):
        self.node = node
        self.style = style
        self.tag = tag
        self.size = size

    @property
    def title(self):
        return titles[self.style]

    def start(self):
        return u"<span class='tag %s' title='%s'>%s" % (self.style,
                                                        self.title,
                                                        self.tag)

    def end(self):
        return self.size, u"</span>"



"""
>>> empty_complex()
0j
>>> new_complex(1., 5)
(1+5j)
>>> convert_to_complex(10)
(10+0j)
>>> convert_to_complex(10+2j)
(10+2j)
>>> convert_to_complex(10.0)
(10+0j)
"""

import sys

from numba import *

@autojit(backend='ast')
def empty_complex():
    x = complex()
    return x

@autojit(backend='ast')
def new_complex(x, y):
    return complex(x, y)

@autojit(backend='ast')
def convert_to_complex(x):
    return complex(x)

if __name__ == '__main__':
    import numba
    numba.testing.testmod()



"""
Compiling @autojit extension classes works as follows:

    * Create an extension Numba/minivect type holding a symtab
    * Capture attribute types in the symtab in the same was as @jit
    * Build attribute hash-based vtable, hashing on (attr_name, attr_type).

        (attr_name, attr_type) is the only allowed key for that attribute
        (i.e. this is fixed at compile time (for now). This means consumers
        will always know the attribute type (and don't need to specialize
        on different attribute types).

        However, using a hash-based attribute table allows easy implementation
        of multiple inheritance (virtual inheritance), without complicated
        C++ dynamic offsets to base objects (see also virtual.py).

    For all methods M with static input types:
        * Compile M
        * Register M in a list of compiled methods

    * Build initial hash-based virtual method table from compiled methods

        * Create pre-hash values for the signatures
            * We use these values to look up methods at runtime

        * Parametrize the virtual method table to build a final hash function:

            slot_index = (((prehash >> table.r) & self.table.m_f) ^
                           self.displacements[prehash & self.table.m_g])

            See also virtual.py and the following SEPs:

                https://github.com/numfocus/sep/blob/master/sep200.rst
                https://github.com/numfocus/sep/blob/master/sep201.rst

            And the following paper to understand the perfect hashing scheme:

                Hash and Displace: Efficient Evaluation of Minimal Perfect
                Hash Functions (1999) by Rasmus Pagn:

                    http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.32.6530

    * Create descriptors that wrap the native attributes
    * Create an extension type:

            {
                hash-based virtual method table (PyCustomSlots_Table **)
                PyGC_HEAD
                PyObject_HEAD
                ...
                native attributes
            }

        We precede the object with the table to make this work in a more
        generic scheme, e.g. where a caller is dealing with an unknown
        object, and we quickly want to see whether it support such a
        perfect-hashing virtual method table:

            if (o->ob_type->tp_flags & NATIVELY_CALLABLE_TABLE) {
                PyCustomSlots_Table ***slot_p = ((char *) o) - sizeof(PyGC_HEAD)
                PyCustomSlots_Table *vtab = **slot_p
                look up function
            } else {
                PyObject_Call(...)
            }

        We need to store a PyCustomSlots_Table ** in the object to allow
        the producer of the table to replace the table with a new table
        for all live objects (e.g. by adding a specialization for
        an autojit method).
"""

import copy
from functools import partial

from numba import error
from numba import pipeline
from numba import typesystem
from numba import numbawrapper
from numba.minivect import minitypes

from numba.exttypes import types as etypes
from numba.exttypes import utils
from numba.exttypes import virtual
from numba.exttypes import signatures
from numba.exttypes import validators
from numba.exttypes import compileclass
from numba.exttypes import ordering
from numba.exttypes import autojitmeta


#------------------------------------------------------------------------
# Build Attributes Struct
#------------------------------------------------------------------------

class AutojitAttributeBuilder(compileclass.AttributeBuilder):

    def finalize(self, ext_type):
        # TODO: hash-based attributes
        ext_type.attribute_table.create_attribute_ordering(ordering.extending)

    def create_descr(self, attr_name):
        """
        Create a descriptor that accesses the attribute on the ctypes struct.
        TODO: Use a perfect-hashed attribute table.
        """
        def _get(self):
            return getattr(self._numba_attrs, attr_name)
        def _set(self, value):
            return setattr(self._numba_attrs, attr_name, value)
        return property(_get, _set)

#------------------------------------------------------------------------
# Filter Methods
#------------------------------------------------------------------------

class AutojitMethodFilter(compileclass.Filterer):

    def filter(self, methods, ext_type):
        typed_methods = []

        for method in methods:
            if method.signature is None:
                # autojit method
                ext_type.vtab_type.untyped_methods[method.name] = method
            else:
                # method with signature
                typed_methods.append(method)

        return typed_methods

#------------------------------------------------------------------------
# Build Method Wrappers
#------------------------------------------------------------------------

class AutojitMethodWrapperBuilder(compileclass.MethodWrapperBuilder):

    def build_method_wrappers(self, env, extclass, ext_type):
        """
        Update the extension class with the function wrappers.
        """
        self.process_typed_methods(env, extclass, ext_type)
        self.process_untyped_methods(env, extclass, ext_type)

    def process_untyped_methods(self, env, extclass, ext_type):
        """
        Process autojit methods (undecorated methods). Use the fast
        NumbaSpecializingWrapper cache when for when we're being called
        from python. When we need to add a new specialization,
        `autojit_method_compiler` is invoked to compile the method.

        extclass: the extension type
        ext_type.py_class: the unspecialized class that was decorated
        """
        from numba.wrapping import compiler

        for method_name, method in ext_type.vtab_type.untyped_methods.iteritems():
            env.specializations.register(method.py_func)
            cache = env.specializations.get_autojit_cache(method.py_func)

            compiler_impl = compiler.MethodCompiler(env, extclass, method)
            wrapper = numbawrapper.NumbaSpecializingWrapper(
                method.py_func, compiler_impl, cache)

            setattr(extclass, method_name, wrapper)

# ______________________________________________________________________
# Compile method when called from Python

def autojit_method_compiler(env, extclass, method, signature):
    """
    Called to compile a new specialized method. The result should be
    added to the perfect hash-based vtable.
    """
    # compiled_method = numba.jit(argtypes=argtypes)(method.py_func)
    func_env = pipeline.compile2(env, method.py_func,
                                 restype=signature.return_type,
                                 argtypes=signature.args)

    # Create Method for the specialization
    new_method = signatures.Method(
        method.py_func,
        method.name,
        func_env.func_signature,
        is_class=method.is_class,
        is_static=method.is_static)

    new_method.update_from_env(func_env)

    # Update vtable type
    vtable_wrapper = extclass.__numba_vtab
    vtable_type = extclass.exttype.vtab_type
    vtable_type.specialized_methods[new_method.name,
                                    signature.args] = new_method

    # Replace vtable (which will update the vtable all (live) objects use)
    new_vtable = virtual.build_hashing_vtab(vtable_type)
    vtable_wrapper.replace_vtable(new_vtable)

    return func_env.numba_wrapper_func

#------------------------------------------------------------------------
# Autojit Extension Class Compiler
#------------------------------------------------------------------------

class AutojitExtensionCompiler(compileclass.ExtensionCompiler):
    """
    Compile @autojit extension classes.
    """

    method_validators = validators.autojit_validators
    exttype_validators = validators.autojit_type_validators

    def get_bases(self):
        """
        Get base classes for the resulting extension type.

        We can try several inheritance schemes. We can choose between:

            1) One unspecialized - general - class

                * This must bind specialized methods on each object instance
                  to allow method calls from Python, making object allocation
                  more expensive

                * We must take the max() of tp_basicsize to allow enough
                  space for all specializations. However, the specialization
                  universe is not known up front, so we must allocate
                  attributes separately on the heap (or perhaps we must
                  override tp_alloc to use a dynamic size depending on the
                  specialization - but this may conflict with non-numba base
                  classes).

            2) One specialized class per instance

                * This allows us to experiment with static layouts for
                  attributes or methods more easily

                * It seems more intuitive to back specialized objects with
                  a specialized type

        We will go with 2). We could go for a specialization tree as follows:

                     A
                   / | \
                A0   |  A1
                 |   |  |
                 |   B  |
                 | /  \ |
                B0     B1

        Which gets us:

            issubclass(A_specialized, A)
            isinstance(A_specialized(), A)

        as well as

            issubclass(B_specialized, A_specialized)

        However, to support this scheme, the unspecialized class A must:

            1) Be subclassable
            2) Return specialized object instances when instantiated
            3) Support unbound method calls

        1) requires that A be a class, and then 2) implies that A's metaclass
        overrides __call__ or that A implements __new__.

        However, since A_specialized subclasses A, A_specialized.__new__ would
        need to skip A.__new__, which requires numba to insert a __new__
        or modify a user's __new__ method in A_specialized.

        The metaclass option seems more feasible:

            A_meta.__call__ -> specialized object instance

        Users can then override a metaclass in a Python (or numba?) subclass
        as follows:

            class MyMeta(type(MyNumbaClass)):
                ...

        The metaclass can also support indexing:

            A_specialized = A[{'attrib_a': double}]
        """
        # TODO: subclassing
        return (self.py_class,)

    def get_metacls(self):
        return autojitmeta.create_specialized_metaclass(self.py_class)

#------------------------------------------------------------------------
# Unbound Methods from Python
#------------------------------------------------------------------------

class UnboundDelegatingMethod(object):
    """
    Function in the unspecialized class that is used for delegation to
    a method in a specialized class, i.e.

        A.method(A(10.0)) -> A(10.0).method()

    This method can never be bound, since __new__ always returns specialized
    instances (so the unspecialized class cannot be instantiated!).
    """

    def __init__(self, py_class, name):
        self.py_class = py_class
        self.name = name

    def __call__(self, obj, *args, **kwargs):
        numbawrapper.unbound_method_type_check(self.py_class, obj)
        return getattr(obj, self.name)(*args, **kwargs)

def make_delegations(py_class):
    """
    Make delegation unbound methods that delegate from the unspecialized
    class to the specialized class. E.g.

        m = A.method
        m(A(10.0))      # Delegate to A[double].method
    """
    class_dict = vars(py_class)
    for name, func in class_dict.iteritems():
        if isinstance(func, (minitypes.Function, staticmethod, classmethod)):
            method = signatures.process_signature(func, name)
            if method.is_class or method.is_static:
                # Class or static method: use the pure Python function wrapped
                # in classmethod()/staticmethod()
                method.wrapper_func = method.py_func
                new_func = method.get_wrapper()
            else:
                # Regular unbound method. Create dispatcher to bound method
                # when called
                new_func = UnboundDelegatingMethod(py_class, name)

            setattr(py_class, name, new_func)

        #------------------------------------------------------------------------
# Make Specializing Class -- Entry Point for decorator application
#------------------------------------------------------------------------

def autojit_class_wrapper(py_class, compiler_impl, cache):
    """
    Invoked when a class is decorated with @autojit.

    :param py_class: decorated python class
    :param compiler_impl: compiler.ClassCompiler
    :param cache: FunctionCache
    :return: py_class that returns specialized object instances
    """
    from numba import numbawrapper

    if utils.is_numba_class(py_class):
        raise error.NumbaError("Subclassing not yet supported "
                               "for autojit classes")

    # runtime_args -> specialized extension type instance
    class_specializer = numbawrapper.NumbaSpecializingWrapper(
        py_class, compiler_impl, cache)

    py_class = autojitmeta.create_unspecialized_cls(py_class, class_specializer)
    py_class.__is_numba_autojit = True

    # Update the class from which we derive specializations (which will be
    # passed to create_extension() below)
    compiler_impl.py_func = py_class

    # Back up class dict, since we're going to modify it
    py_class.__numba_class_dict = dict(vars(py_class))

    # Make delegation methods for unbound methods
    make_delegations(py_class)

    # Setup up partial compilation environment
    # partial_env = create_partial_extension_environment(
    #     compiler_impl.env, py_class, compiler_impl.flags)
    # compiler_impl.partial_ext_env = partial_env

    return py_class

#------------------------------------------------------------------------
# Build Extension Type -- Compiler Entry Point
#------------------------------------------------------------------------

# def create_partial_extension_environment(env, py_class, flags, argtypes):
def create_extension(env, py_class, flags, argtypes):
    """
    Create a partial environment to compile specialized versions of the
    extension class in.

    Inovked when calling the wrapped class to compile a specialized
    new extension type.
    """
    # TODO: Remove argtypes! Partial environment!

    flags.pop('llvm_module', None)

    ext_type = etypes.AutojitExtensionType(py_class)

    class_dict = dict(utils.get_class_dict(py_class))

    extension_compiler = AutojitExtensionCompiler(
        env, py_class, class_dict, ext_type, flags,
        signatures.AutojitMethodMaker(argtypes),
        compileclass.AttributesInheriter(),
        AutojitMethodFilter(),
        AutojitAttributeBuilder(),
        virtual.HashBasedVTabBuilder(),
        AutojitMethodWrapperBuilder())

    extension_compiler.init()
    # return extension_compiler

# def compile_class(extension_compiler, argtypes):
    extension_compiler.infer()
    extension_compiler.finalize_tables()
    extension_compiler.validate()
    extension_type = extension_compiler.compile()

    return extension_type


#! /usr/bin/env python
# ______________________________________________________________________
'''test_type_inference

Test type inference.
'''
# ______________________________________________________________________

from numba import *
from numba import typesystem
from numba import decorators

import unittest

import numpy
import numpy as np

import logging

logging.basicConfig(level=logging.DEBUG)
# ______________________________________________________________________

def _simple_func(arg):
    if arg > 0.:
        result = 22.
    else:
        result = 42.
    return result

simple_func = decorators.autojit(backend='ast')(_simple_func)

def _for_loop(start, stop, inc):
    acc = 0
    for value in range(start, stop, inc):
        acc += value

    print(value)
    return acc

for_loop = decorators.autojit(backend='ast')(_for_loop)

def arange():
    a = numpy.arange(10)
    b = numpy.arange(10, dtype=numpy.double)
    return a, b

def empty_like(a):
    b = numpy.empty_like(a)
    c = numpy.zeros_like(a, dtype=numpy.int32)
    d = numpy.ones_like(a)

dtype = np.float32

def _empty(N):
    # default dtype
    a1 = numpy.empty(N)
    a2 = numpy.empty((N,))
    a3 = numpy.empty([N])

    # Given dtype
    a4 = numpy.empty(N, dtype)
    a5 = numpy.empty(N, dtype=dtype)
    a6 = numpy.empty(N, np.float32)
    a7 = numpy.empty(N, dtype=np.float32)

    # Test dimensionality
    a8 = np.empty((N, N), dtype=np.int64)

def _empty_arg(N, empty, zeros, ones):
    a1 = empty([N])
    a2 = zeros([N])
    a3 = ones([N])

@autojit
def assert_array_dtype(A, value, empty, zeros, ones):
    if value < 2:
        A = empty([A.shape[0], A.shape[1]], dtype=A.dtype)
    elif value < 4:
        A = zeros([A.shape[0], A.shape[1]], dtype=A.dtype)
    elif value < 6:
        A = ones([A.shape[0], A.shape[1]], dtype=A.dtype)
    else:
        pass

    # 'A' must have a consistent array type here
    return A

def slicing(a):
    n = numpy.newaxis

    # 0D
    b = a[0]
    c = a[9]

    # 1D
    d = a[:]
    e = a[...]

    # 2D
    f = a[n, ...]
    g = a[numpy.newaxis, :]
    h = a[..., numpy.newaxis]
    i = a[:, n]

    j = a[n, numpy.newaxis, 0]
    k = a[numpy.newaxis, 0, n]
    l = a[0, n, n]

def none_newaxis(a):
    n = None

    # 2D
    f = a[None, ...]
    #g = a[n, :]
    h = a[..., None]
    #i = a[:, n]

    #j = a[n, None, 0]
    k = a[None, 0, numpy.newaxis]
    l = a[0, n, numpy.newaxis]

def func_with_signature(a):
    if a > 1:
        return float(a)
    elif a < 5:
        return int(a)
    elif a > 10:
        return object()

    return a + 1j

def arg_rebind(a):
    a = 0
    a = 0.0
    a = "hello"

# ______________________________________________________________________

from numba.control_flow.tests.test_cfg_type_infer import infer, functype

class TestTypeInference(unittest.TestCase):
    def test_simple_func(self):
         self.assertEqual(simple_func(-1.), 42.)
         self.assertEqual(simple_func(1.), 22.)

    def test_simple_for(self):
         self.assertEqual(for_loop(0, 10, 1), 45)

    def test_type_infer_simple_func(self):
        sig, symtab = infer(_simple_func, functype(None, [double]))
        self.assertEqual(sig.return_type, double)

    def test_type_infer_for_loop(self):
        sig, symtab = infer(_for_loop, functype(None, [int_, int_, int_]))
        self.assertTrue(symtab['acc'].type.is_int)
        self.assertEqual(symtab['value'].type, Py_ssize_t)
        self.assertEqual(sig.return_type, Py_ssize_t)

    def test_type_infer_arange(self):
        sig, symtab = infer(arange, functype())
        self.assertEqual(symtab['a'].type, npy_intp[:])
        self.assertEqual(symtab['b'].type, double[:])

    def test_empty_like(self):
        sig, symtab = infer(empty_like, functype(None, [double[:]]))
        self.assertEqual(symtab['b'].type, double[:])
        self.assertEqual(symtab['c'].type, int32[:])
        self.assertEqual(symtab['d'].type, double[:])

    def test_empty(self):
        sig, symtab = infer(_empty, functype(None, [int_]))
        for i in range(1, 4):
            self.assertEqual(symtab['a%d' % i].type, double[:])

        for i in range(4, 8):
            self.assertEqual(symtab['a%d' % i].type, float_[:])

        self.assertEqual(symtab['a8'].type, int64[:, :])

    def test_empty_arg(self):
        from numba import typesystem as nt
        empty_t = nt.ModuleAttributeType(module=np, attr='empty')
        zeros_t = nt.ModuleAttributeType(module=np, attr='zeros')
        ones_t = nt.ModuleAttributeType(module=np, attr='ones')

        sig, symtab = infer(_empty_arg, functype(None, [int_, empty_t,
                                                        zeros_t, ones_t]))
        for i in range(1, 4):
            self.assertEqual(symtab['a%d' % i].type, double[:])

    def test_dtype_attribute(self):
        A = np.empty((10, 10), dtype=np.float32)

        A_result = assert_array_dtype(A, 3, np.empty, np.zeros, np.ones)
        assert np.all(A_result == 0)
        A_result = assert_array_dtype(A, 5, np.empty, np.zeros, np.ones)
        assert np.all(A_result == 1)

    def test_slicing(self):
        sig, symtab = infer(slicing, functype(None, [double[:]]))
        self.assertEqual(symtab['n'].type, typesystem.NewAxisType())

        self.assertEqual(symtab['b'].type, double)
        self.assertEqual(symtab['c'].type, double)
        self.assertEqual(symtab['d'].type, double[:])
        self.assertEqual(symtab['e'].type, double[:])
        self.assertEqual(symtab['f'].type, double[:, :])
        self.assertEqual(symtab['g'].type, double[:, :])
        self.assertEqual(symtab['h'].type, double[:, :])
        self.assertEqual(symtab['i'].type, double[:, :])
        self.assertEqual(symtab['j'].type, double[:, :])
        self.assertEqual(symtab['k'].type, double[:, :])
        self.assertEqual(symtab['l'].type, double[:, :])

    def test_none_newaxis(self):
        sig, symtab = infer(none_newaxis, functype(None, [double[:]]))
        self.assertEqual(symtab['f'].type, double[:, :])
        #self.assertEqual(symtab['g'].type, double[:, :])
        self.assertEqual(symtab['h'].type, double[:, :])
        #self.assertEqual(symtab['i'].type, double[:, :])
        #self.assertEqual(symtab['j'].type, double[:, :, :])
        self.assertEqual(symtab['k'].type, double[:, :])
        self.assertEqual(symtab['l'].type, double[:, :])

    def test_return_type(self):
        sig, symtab = infer(func_with_signature, functype(int_, [int_]))
        assert sig == int_(int_)

        sig, symtab = infer(func_with_signature, functype(int_, [float_]))
        assert sig == int_(float_)

        sig, symtab = infer(func_with_signature, functype(float_, [int_]))
        assert sig == float_(int_)

    def test_rebind_arg(self):
        sig, symtab = infer(arg_rebind, functype(int_, [int_]),
                            allow_rebind_args=True)
        assert sig == int_(int_)
        assert symtab['a'].type == c_string_type

#        try:
#            sig, symtab = infer(arg_rebind, functype(int_, [int_]),
#                                allow_rebind_args=False)
#        except minierror.UnpromotableTypeError as e:
#            msg = str(sorted(e.args, key=str))
#            self.assertEqual("[(double, const char *)]", msg)
#        else:
#            raise Exception("Expected an unpromotable type error")

# ______________________________________________________________________

if __name__ == "__main__":
#    TestTypeInference('test_dtype_attribute').debug()
    unittest.main()
